from typing import Optional
from nicegui import ui, events
from ..config import logger


class TypeView:

    colors = {'item': 'bg-stone-100',
              'cls_color': 'bg-stone-300',
              'selected': 'bg-blue-200'}

    detail_view = None

    def __init__(self, *args, **kwargs) -> None:

        self._component = None
        self.checkbox = None
        self.component = kwargs.get('component', None)
        self.parent = kwargs.get('parent', None)

        self.row = None
        self.card: Optional[ui.card, None] = None

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, value):
        self._component = value

        try:
            self._component.__ui_element__ = self
        except Exception as e:
            logger.error(f'Error setting __ui_element__ attribute: {e}')

    @property
    def selected(self):
        if self.checkbox is None:
            return False
        return self.checkbox.value

    @ui.refreshable
    def ui_content(self):

        from .detail_views import show_detail

        with ui.list().classes('w-full h-full') as self.card:
        # with ui.card().classes('w-full h-full').props('color="blue-800" keep-color') as self.card:
            with ui.item().classes('w-full h-full'):
                with ui.item_section():
                    self.checkbox = ui.checkbox()
                with ui.item_section():
                    ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')
                with ui.item_section():
                    ui.label(f'{str(self.component.id)}')
                with ui.item_section():
                    ui.button(on_click=lambda e: show_detail(value=self.component,
                                                             parent=self), icon='open')

    def select(self, e: events.ClickEventArguments):
        if self.selected:
            self.row.classes(remove=f"{self.colors['item']}", add=f"{self.colors['selected']}")
            self.card.classes(remove=f"{self.colors['item']}", add=f"{self.colors['selected']}")
        else:
            self.row.classes(add=f"{self.colors['item']}", remove=f"{self.colors['selected']}")
            self.card.classes(add=f"{self.colors['item']}", remove=f"{self.colors['selected']}")



import os
from nicegui import app as ng_app, ui, Client
from . import user_manager


from .app.theme import frame

from .app.auth import AuthMiddleware
from .app import login

unrestricted_page_routes = {'/login'}


# ui.add_head_html("<style>" + open(Path(__file__).parent / "static_files" / "styles.css").read() + "</style>")


@ui.page('/')
def index_page(client: Client) -> None:
    frame('Home')
    login.content()
    # with frame('Home'):
    #     home.content()


@ui.page('/login')
def index_page1() -> None:
    # frame('Home')
    # with frame('Home'):
    #     pass
    login.content()


@ui.page('/project')
def index_page2(client: Client) -> None:
    # frame('Home')
    # with frame('Home'):
    #     pass
    frame('Project')


# def on_startup():
#
#
#
# ng_app.on_startup(on_startup)


# app.on_shutdown(handle_shutdown)
def run_ui(*args, **kwargs):
    ng_app.add_middleware(AuthMiddleware)
    storage_secret = kwargs.get('storage_secret', os.environ.get('STORAGE_SECRET', 'my_secret6849'))

    # set_storage_secret(storage_secret)
    ui.run(storage_secret=storage_secret,
           title=kwargs.get('title', 'Py Simultan'),
           dark=kwargs.get('dark', False),
           reload=kwargs.get('reload', True),
           port=kwargs.get('port', 8080),
           uvicorn_logging_level=kwargs.get('uvicorn_logging_level', 'info'),
           )
    # set_storage_secret(storage_secret)

    from .core.patch_default_models import patch
    from . import user_manager
    patch(user_manager)


if __name__ == '__main__':
    run_ui()



import shutil
import os
from typing import Optional
from nicegui import ui, events, app

from .type_view_manager import TypeViewManager
from .. import user_manager

from SIMULTAN.Data.MultiValues import (SimMultiValueField3D, SimMultiValueField3DParameterSource, SimMultiValueBigTable,
                                       SimMultiValueBigTableHeader, SimMultiValueBigTableParameterSource)


class ArrayManager(TypeViewManager):

    columns = [{'name': 'id',
                'label': 'ID',
                'field': 'id',
                'sortable': True,
                'align': 'left'},
               {'name': 'name',
                'label': 'Name',
                'field': 'name',
                'sortable': True,
                'align': 'left'},
               {'name': 'shape',
                'label': 'Shape',
                'field': 'shape',
                'sortable': False,
                'align': 'left'},
               ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.table: Optional[ui.table] = None

        self.df_items = {}
        self.np_items = {}

    @property
    def data_model(self):
        return self._data_model

    @data_model.setter
    def data_model(self, value):
        self._data_model = value
        self.df_items = self.update_df_items()
        self.np_items = self.update_np_items()
        self.ui_content.refresh()

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @property
    def mapper(self):
        return self.user.mapper

    @property
    def project_manager(self):
        return self.user.project_manager

    def update_df_items(self) -> list[str: any]:
        if self.data_model is None:
            return {}
        return {str(x.Id): x for x in self.data_model.value_fields if isinstance(x, SimMultiValueBigTable)}

    def update_np_items(self) -> dict[str: any]:
        if self.data_model is None:
            return {}
        return {str(x.Id): x for x in self.data_model.value_fields if isinstance(x, SimMultiValueField3D)}

    def button_create_table_ui_content(self):
        ui.button('Create new table', on_click=self.create_new_table)

    def button_create_np_ui_content(self):
        ui.button('Create new array', on_click=self.create_new_array)

    @ui.refreshable
    def ui_content(self):
        self.ui_df_content()
        self.ui_np_content()

    def ui_df_content(self):
        df_items = self.update_df_items()

        rows = [{'id': f'{item.Id}',
                 'name': item.Name,
                 'shape': f'{(item.RowHeaders.Count, item.ColumnHeaders.Count)}'
                 }
                for item in df_items.values()]

        with ui.table(columns=self.columns,
                      rows=rows,
                      title='Tables',
                      selection='single',
                      row_key='id').classes('w-full bordered') as self.table:

            self.table.add_slot('body', r'''
                                <q-tr :props="props">
                                    <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                        {{ col.value }}
                                    </q-td>
                                    <q-td auto-width>
                                        <q-btn size="sm" color="blue" round dense
                                               @click="$parent.$emit('show_detail', props)"
                                               icon="launch" />
                                    </q-td>
                                </q-tr>
                            ''')

            self.table.on('show_detail', self.show_df_detail)

        self.button_create_table_ui_content()

    def ui_np_content(self):
        df_items = self.update_np_items()

        rows = [{'id': f'{item.Id}',
                 'name': item.Name,
                 'shape': f'{(len(item.XAxis), len(item.YAxis), len(item.ZAxis))}'
                 }
                for item in df_items.values()]

        with ui.table(columns=self.columns,
                      rows=rows,
                      title='Arrays',
                      selection='single',
                      row_key='id').classes('w-full bordered') as self.table:

            self.table.add_slot('body', r'''
                                <q-tr :props="props">
                                    <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                        {{ col.value }}
                                    </q-td>
                                    <q-td auto-width>
                                        <q-btn size="sm" color="blue" round dense
                                               @click="$parent.$emit('show_detail', props)"
                                               icon="launch" />
                                    </q-td>
                                </q-tr>
                            ''')

            self.table.on('show_detail', self.show_np_detail)

        self.button_create_np_ui_content()

    def show_df_detail(self, e: events.GenericEventArguments):
        from .detail_views import show_detail
        table: SimMultiValueBigTable = self.df_items.get(e.args['row']['id'], None)
        show_detail(table)

    def show_np_detail(self, e: events.GenericEventArguments):
        from .detail_views import show_detail
        array: SimMultiValueBigTable = self.np_items.get(e.args['row']['id'], None)
        show_detail(array)

    def add_df_item_to_view(self, item: SimMultiValueBigTable):

        if str(item.Id) not in (str(x['id']) for x in self.table.rows):
            self.table.add_rows({'id': f'{item.Id}',
                                 'name': item.Name,
                                 'shape': f'{(item.RowHeaders.Count, item.ColumnHeaders.Count)}'
                                 })

    def add_np_item_to_view(self, item: SimMultiValueBigTable):

        if str(item.Id) not in (x['id'] for x in self.table.rows):
            self.table.add_rows({'id': f'{item.Id}',
                                 'name': item.Name,
                                 'shape': f'{(len(item.XAxis), len(item.YAxis), len(item.ZAxis))}'
                                 })

    def create_new_table(self):
        ui.notify('Create new table not implemented yet!', type='negative')

    def create_new_array(self):
        ui.notify('Create new array not implemented yet!', type='negative')



from nicegui import Client, app, ui
# from .theme import frame
# from ..router import router
from typing import Optional
from fastapi.responses import RedirectResponse

from .. import user_manager


def content() -> Optional[RedirectResponse]:
    def try_login() -> None:  # local function to avoid passing username and password as arguments
        if user_manager.authenticate(username.value, password.value):
            app.storage.user.update({'username': username.value, 'authenticated': True})
            ui.open(app.storage.user.get('referrer_path', '/'))  # go back to where the user wanted to go
            logger = user_manager.users[app.storage.user['username']].logger
            if logger is not None:
                user_manager.users[app.storage.user['username']].logger.info('Logged in')
        else:
            ui.notify('Wrong username or password', color='negative')

    if app.storage.user.get('authenticated', False):
        return RedirectResponse('/')
    with ui.card().classes('h-full w-full').style('align-items: center;'):
        ui.label('PySimultan WebUI')
        with ui.card():
            username = ui.input('Username').on('keydown.enter', try_login)
            password = ui.input('Password', password=True, password_toggle_button=True).on('keydown.enter', try_login)
            ui.button('Log in', on_click=try_login)



import os
import sys
import subprocess
import platform

from logging import getLogger

logger = getLogger('py_simultan_ui')


def find_freecad_path_registry():

    import winreg

    try:
        # Open the registry key
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall")
        i = 0
        while True:
            try:
                subkey_name = winreg.EnumKey(key, i)
                subkey = winreg.OpenKey(key, subkey_name)
                try:
                    # Check for a display name entry in the subkey
                    name, _ = winreg.QueryValueEx(subkey, "DisplayName")
                    # Check if it's FreeCAD
                    if "FreeCAD" in name:
                        # Get the installation path
                        path, _ = winreg.QueryValueEx(subkey, "DisplayIcon")
                        return os.path.dirname(path)
                except FileNotFoundError:
                    pass
                finally:
                    subkey.Close()
                i += 1
            except WindowsError:
                break
    except Exception as e:
        return
    return


def find_freecad_path():
    try:
        logger.info('Searching for FreeCAD executable')

        # Use the 'where' command to find FreeCAD executable
        result = subprocess.run("where /R C:\ freecad.exe", capture_output=True, text=True, timeout=15)
        if result.returncode == 0:
            # Successful, extract the path
            paths = result.stdout.split('\n')
            # Usually the first result is the most relevant one
            return os.path.dirname(paths[0]) if paths else None
        else:
            return
    except Exception as e:
        logger.error(f'Could not find FreeCAD executable')
        return


if platform.system() == 'Windows':

    FREECADPATH = os.environ.get('FREECADPATH', None)
    if FREECADPATH is None:
        if os.path.isdir(r'C:\Program Files'):
            if 'FreeCAD' in os.listdir(r'C:\Program Files'):
                freecad_versions = os.listdir(r'C:\Program Files\FreeCAD')
                freecad_versions.sort(reverse=True)
                FREECADPATH = os.path.join(r'C:\Program Files\FreeCAD', freecad_versions[0], 'bin')
        if os.path.isdir(r'D:\Program Files'):
            if 'FreeCAD' in os.listdir(r'D:\Program Files'):
                freecad_versions = os.listdir(r'D:\Program Files\FreeCAD')
                freecad_versions.sort(reverse=True)
                FREECADPATH = os.path.join(r'D:\Program Files\FreeCAD', freecad_versions[0], 'bin')

                sys.path.append(os.path.join(r'D:\Program Files\FreeCAD', freecad_versions[0]))

    if FREECADPATH is None:
        FREECADPATH = find_freecad_path_registry()
    if FREECADPATH is None:
        FREECADPATH = find_freecad_path()

    if FREECADPATH is not None:
        sys.path.append(FREECADPATH)



from PySimultan2 import Content

contents = {}


contents['param_a'] = Content(text_or_key='param_a',
                              property_name='param_a',
                              type=None,
                              unit=None,
                              documentation='param_a')

contents['param_b'] = Content(text_or_key='param_b',
                              property_name='param_b',
                              type=None,
                              unit=None,
                              documentation='param_b')

contents['param_c'] = Content(text_or_key='param_c',
                              property_name='param_c',
                              type=None,
                              unit=None,
                              documentation='param_c')

contents['param_d'] = Content(text_or_key='param_d',
                              property_name='param_d',
                              type=None,
                              unit=None,
                              documentation='param_d')

contents['thermal_conductivity'] = Content(text_or_key='thermal_conductivity',
                                           property_name='thermal_conductivity',
                                           type=None,
                                           unit=None,
                                           documentation='thermal_conductivity')

contents['thermal_resistance'] = Content(text_or_key='thermal_resistance',
                                         property_name='thermal_resistance',
                                         type=None,
                                         unit=None,
                                         documentation='thermal_resistance')



from nicegui import ui
from PySimultan2.default_types import ComponentList
from ..type_view_manager import TypeViewManager


class ComponentListManager(TypeViewManager):

    cls = ComponentList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def button_create_ui_content(self):
        ui.button('Create new Component list', on_click=self.create_new_item, icon='add')

    def create_new_item(self, event):
        if self.data_model is None:
            ui.notify('No data model selected! Please load a data model first.')
            return

        new_item = self.cls(data_model=self.data_model)
        self.add_item_to_view(new_item)



import os

from PySimultanUI.src.pysimultanui import user_manager, run_ui
from pysimultan_freecad.src.pysimultan_freecad import mapper, method_mapper, view_manager

project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)

mapping = user_manager.mapper_manager.create_mapping(name='FreeCAD Toolbox')
mapping.mapper = mapper
mapping.method_mapper = method_mapper
mapping.view_manager = view_manager

run_ui()
print('Done!')



import pandas as pd
import asyncio
import numpy as np
from logging import getLogger
from copy import copy
from nicegui import ui, run, app
from typing import Optional, TypedDict

from .type_view_manager import TypeViewManager
from .type_view import TypeView
from .. import core
from .. import user_manager
from ..core import PythonMapper

from .mapped_cls.mapped_cls_manager import MappedClsManager
from .mapped_cls.mapped_cls_view import MappedClsView

from .parameter_view import ParameterView
from .component_list_view.component_list_manager import ComponentListManager
from .component_list_view.component_list_view import ComponentListView

from .component_dict_view import ComponentDictManager, ComponentDictView

from .pandas_df_view import DataFrameView, DataFrameManager
from .numpy_view import NDArrayView, NDArrayManager

from PySimultan2.simultan_object import SimultanObject


logger = getLogger('py_simultan_ui')


class ClassView(TypedDict, total=False):
    item_view_manager: Optional[TypeViewManager]
    type_view: TypeView


class ViewManager(object):

    def __init__(self, *args, **kwargs):

        self._data_model = None
        self._mapper = None
        self.mapper = kwargs.get('mapper', core.mapper)
        self.cls_views: dict[any, ClassView] = {}
        self.data_model = kwargs.get('data_model', None)
        self.parent = kwargs.get('parent', None)

        self.cls_views[int] = {'item_view_manager': None, 'type_view': ParameterView}
        self.cls_views[float] = {'item_view_manager': None, 'type_view': ParameterView}
        self.cls_views[str] = {'item_view_manager': None, 'type_view': ParameterView}

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @property
    def mapper(self):
        return self._mapper

    @mapper.setter
    def mapper(self, value):
        self._mapper = value
        self.cls_views = {}
        self.create_tvms()

    def register(self,
                 cls: type,
                 type_view: type[TypeView],
                 item_view_manager: Optional[TypeViewManager] = None,):

        self.cls_views[cls] = {'item_view_manager': item_view_manager, 'type_view': type_view}
        cls.item_view_manager = item_view_manager
        cls.type_view = type_view

    @property
    def data_model(self):
        return self._data_model

    @data_model.setter
    def data_model(self, value):

        logger.debug(f'View Manager setting data model to {value}')

        self._data_model = value

        for cls, cls_view_dict in self.cls_views.items():
            item_view_manager = cls_view_dict.get('item_view_manager', None)
            if item_view_manager is None:
                continue
            else:
                item_view_manager.data_model = value

    def get_view(self, cls: type):
        return self.cls_views[cls]

    def create_mapped_cls_view_manager(self, taxonomy: str) -> Optional[TypeViewManager]:
        cls = self.mapper.get_mapped_class(taxonomy)

        if self.cls_views.get(cls, None) is None:
            new_tvm = MappedClsManager(mapper=self.mapper)
            new_tvm.cls = self.mapper.get_mapped_class(taxonomy)
            # old__init__ = copy(new_tvm.cls.__init__)
            #
            # def __init__(self, *args, **kwargs):
            #     old__init__(self, *args, **kwargs)
            #     new_tvm.add_item_to_view(self)
            #
            # new_tvm.cls.__init__ = __init__

            new_tvm.item_view_cls = MappedClsView
            new_tvm.item_view_name = taxonomy
            new_tvm.view_manager = self
            self.cls_views[new_tvm.cls] = {'item_view_manager': new_tvm, 'type_view': MappedClsView}

            with self.user.project_manager.project_tab:
                new_tvm.ui_content()

        return self.cls_views.get(cls, None)

    def create_tvms(self):

        logger.debug('Creating TypeViewManagers...')

        def add_new_init_method(tvm: TypeViewManager):
            old__init__ = copy(tvm.cls.__init__)

            def __init__(self, *args, **kwargs):
                old__init__(self, *args, **kwargs)
                tvm.add_item_to_view(self)

            tvm.cls.__init__ = __init__

        for taxonomy, cls in self.mapper.registered_classes.items():

            mapped_cls = self.mapper.get_mapped_class(taxonomy)

            if taxonomy == 'ComponentList':
                new_tvm = ComponentListManager(mapper=self.mapper)
                new_tvm.item_view_cls = ComponentListView

            elif taxonomy == 'ComponentDict':
                new_tvm = ComponentDictManager(mapper=self.mapper)
                new_tvm.item_view_cls = ComponentDictView

            elif self.cls_views.get(cls, None) is None:
                new_tvm = MappedClsManager(mapper=self.mapper)
                new_tvm.item_view_cls = MappedClsView

            else:
                new_tvm = self.cls_views.get(cls, None)['item_view_manager']

            if new_tvm is None:
                continue

            new_tvm.cls = mapped_cls
            # add_new_init_method(new_tvm)
            new_tvm.item_view_name = taxonomy
            new_tvm.view_manager = self
            self.cls_views[new_tvm.cls] = {'item_view_manager': new_tvm, 'type_view': new_tvm.item_view_cls}

        df_manager = DataFrameManager(mapper=self.mapper)
        self.cls_views[pd.DataFrame] = {'item_view_manager': df_manager, 'type_view': DataFrameView}

        np_manager = NDArrayManager(mapper=self.mapper)
        self.cls_views[np.ndarray] = {'item_view_manager': np_manager, 'type_view': NDArrayView}

    @ui.refreshable
    def ui_content(self):

        if len(self.cls_views) < len(self.mapper.registered_classes):
            self.create_tvms()

        logger.debug('Creating UI content')

        self.user.method_mapper.ui_content()
        for tvm_dict in self.cls_views.values():
            tvm: Optional[TypeViewManager, None] = tvm_dict.get('item_view_manager', None)
            if tvm is None:
                continue
            tvm.ui_content()

    def refresh_all_items(self):

        logger.debug('Refreshing all items...')

        n = ui.notification(timeout=None)
        n.spinner = True
        n.message = 'Updating all items...'

        for tvm_dict in self.cls_views.values():
            tvm: Optional[TypeViewManager, None] = tvm_dict.get('item_view_manager', None)
            if tvm is None:
                continue
            print(f'Updating items of {tvm.item_view_name}...')
            n.message = f'Updating items of {tvm.item_view_name}...'
            tvm.update_items_views()

        ui.notify('All items updated!', type='positive')

        n.message = 'Updating all items done!'
        n.type = 'positive'
        n.spinner = False

        # self.method(*args, **kwargs)
        n.dismiss()



from typing import Optional

from .. import core
from ..core.method_mapper import method_mapper
from nicegui import Client, app, ui, events
from PySimultan2.simultan_object import SimultanObject

from ..views import view_manager


def content():
    pass


def static_content():
    pass



import sys
import traceback
from typing import Any, Optional, List, Union
from .scene import ExtendedScene
import numpy as np
from ..core import freecad_utils

import logging

logger = logging.getLogger('PySimultanUI')

import FreeCAD
import Part as FCPart


def rgb_to_hex(rgb: List[int]) -> str:
    if rgb is None:
        return '#000000'
    return f'#{rgb[0] << 16 | rgb[1] << 8 | rgb[2]:06x}'


def display_vertices(vertices: List[FCPart.Vertex],
                     scene: ExtendedScene,
                     obj_id: str = '',
                     colors: list[list[int]] = None):

    try:
        if colors is None:
            colors = [[0, 0, 0]] * vertices.__len__()

        point_cloud = scene.point_cloud([list(x.Point) for x in vertices],
                                        colors=colors,
                                        point_size=0.1).with_name(obj_id)
    except Exception as e:
        logger.error(f'Error displaying vertices: {e}\n'
                     f'{traceback.format_exception(*sys.exc_info())}')
        point_cloud = None

    return point_cloud


def display_edges(edges: Union[List[FCPart.Edge], FCPart.Edge],
                  scene: ExtendedScene,
                  obj_id: str = '',
                  color: Optional[list[int]] = None):

    rgb_color = rgb_to_hex(color)

    if isinstance(edges, FCPart.Edge):
        edges = [edges]

    objects3d = []

    for edge in edges:
        try:
            if edge.Curve.TypeId == 'Part::GeomLine':
                objects3d.append(scene.line(list(edge.Vertexes[0].Point),
                                            list(edge.Vertexes[1].Point)).material(rgb_color).with_name(obj_id))
            else:

                n_points = int(edge.Length / 100) + 1
                if n_points < 100:
                    n_points = 100
                elif n_points > 1000:
                    n_points = 1000

                wire_vertices = np.array(list(edge.discretize(n_points)))
                objects3d.append(scene.wire(wire_vertices).material(rgb_color).with_name(obj_id))
        except Exception as e:
            logger.error(f'Error displaying edges: {e}\n'
                         f'{traceback.format_exception(*sys.exc_info())}')

    return objects3d


def display_wire(wire: FCPart.Wire,
                 scene: ExtendedScene,
                 obj_id: str = '',
                 color: Optional[list[int]] = None):

    try:
        return display_edges(wire.Edges, scene, obj_id, color)
    except Exception as e:
        logger.error(f'Error displaying wire: {e}\n'
                     f'{traceback.format_exception(*sys.exc_info())}')


def display_face(face: FCPart.Face,
                 scene: ExtendedScene,
                 obj_id: str = '',
                 color: Optional[list[int]] = None):

    try:
        rgb_color = rgb_to_hex(color)
        tris = face.tessellate(5, True)
        return scene.mesh(np.array(tris[0]),
                          np.array(tris[1])
                          ).material(rgb_color).with_name(obj_id)
    except Exception as e:
        logger.error(f'Error displaying face: {e}\n'
                     f'{traceback.format_exception(*sys.exc_info())}')


def display_solid(solid: FCPart.Solid,
                  scene: ExtendedScene,
                  obj_id: str = '',
                  color: Optional[list[int]] = None):
    try:
        rgb_color = rgb_to_hex(color)
        tris = solid.tessellate(5, True)

        return scene.mesh(np.array(tris[0]),
                          np.array(tris[1])
                          ).material(rgb_color).with_name(obj_id)
    except Exception as e:
        logger.error(f'Error displaying solid: {e}\n'
                     f'{traceback.format_exception(*sys.exc_info())}')



import os
import io
import shutil
import traceback
import sys
import importlib
import logging
from nicegui import ui, events, app
from PySimultan2.object_mapper import PythonMapper
from .method_mapper import MethodMapper

from typing import TYPE_CHECKING, Any, Optional
if TYPE_CHECKING:
    from .user import User
    from ..views.detail_views import DetailView

logger = logging.getLogger('PySimultanUI')
logger.info('Imported mappers')


project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)

app.add_static_files('/project', project_dir)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ViewManager:

    def __init__(self, *args, **kwargs):
        self.views: dict[str, 'DetailView'] = {}

    def add_view(self,
                 taxonomy_entry_key: str,
                 view: 'DetailView'):
        self.views[taxonomy_entry_key] = view


class Mapping:

    def __init__(self,
                 user: 'User',
                 mapper: 'PythonMapper',
                 method_mapper: 'MethodMapper',
                 view_manager: 'ViewManager' = None,
                 name: str = 'Mapping'):

        self._method_mapper = None
        self._mapper = None

        self.name: str = name
        self.mapper: 'PythonMapper' = mapper
        self.method_mapper: 'MethodMapper' = method_mapper
        if self.method_mapper.mapper is None:
            self.method_mapper.mapper = self.mapper

        self.view_manager: ViewManager = view_manager if view_manager is not None else ViewManager()
        self.user: 'User' = user

    @property
    def method_mapper(self) -> 'MethodMapper':
        return self._method_mapper

    @method_mapper.setter
    def method_mapper(self, value: 'MethodMapper') -> None:
        self._method_mapper = value
        if self._method_mapper is not None:
            if self._method_mapper.mapper is None:
                self._method_mapper.mapper = self.mapper

    @property
    def mapper(self) -> 'PythonMapper':
        return self._mapper

    @mapper.setter
    def mapper(self, value: 'PythonMapper') -> None:
        self._mapper = value
        if self._method_mapper is not None:
            if self._method_mapper.mapper is None:
                self._method_mapper.mapper = self.mapper

    def create_mapped_classes(self, *args, **kwargs):
        _ = [self.mapper.get_mapped_class(x) for x in self.mapper.registered_classes.keys()]

    def copy(self, user: 'User'):

        new_method_mapper = self.method_mapper.copy() if self.method_mapper is not None else None
        new_mapper = new_method_mapper.mapper.copy() if new_method_mapper.mapper is not None else None
        new_mapper.method_mapper = new_method_mapper
        new_mapper.clear()

        _ = [new_mapper.get_mapped_class(x) for x in new_mapper.registered_classes.keys()]

        return Mapping(mapper=new_mapper,
                       user=user,
                       method_mapper=new_method_mapper,
                       view_manager=self.view_manager,
                       name=self.name)


class MapperManager(metaclass=Singleton):

    def __init__(self,
                 available_mappings: dict[str: Mapping] = None,
                 mappings: dict['User': list[Mapping]] = None) -> None:

        self.available_mappings: dict[str: Mapping] = available_mappings if available_mappings is not None else {}
        self.mappings: dict['User':
                            list[Mapping]] = mappings if mappings is not None else {}

        self.create_mapping(name='Default')
        logger.info(f'MapperManager initialized {id(self)}')

    def create_mapping(self,
                       name: str = 'Mapping',
                       user: 'User' = None,
                       mapper: 'PythonMapper' = None,
                       method_mapper: 'MethodMapper' = None,
                       view_manager: ViewManager = None) -> Mapping:

        if name in self.available_mappings:
            ui.notify(f'Mapping with name {name} already exists')
            raise ValueError(f'Mapping with name {name} already exists')

        if mapper is None:
            mapper = PythonMapper()
            _ = [mapper.get_mapped_class(x) for x in mapper.registered_classes.keys()]

        if method_mapper is None:
            method_mapper = MethodMapper(mapper=mapper)

        mapping = Mapping(mapper=mapper,
                          user=user,
                          name=name,
                          method_mapper=method_mapper,
                          view_manager=view_manager
                          )

        self.available_mappings[name] = mapping
        return mapping

    def get_mapping(self,
                    user: 'User',
                    name: str,
                    load_undefined: bool = True
                    ) -> Mapping:

        """
        Get a mapping for a user by name. If the mapping does not exist, create it.
        :param user:
        :param name:
        :param load_undefined: load undefined components
        :return:
        """

        if name is None:
            return None

        if self.mappings.get(user, None) is None:
            self.mappings[user] = {}

        if self.mappings[user].get(name) is None:
            self.mappings[user][name] = self.available_mappings[name].copy(user=user)

        mapping = self.mappings[user][name]
        mapping.mapper.load_undefined = load_undefined

        _ = [mapping.mapper.get_mapped_class(x) for x in mapping.mapper.registered_classes.keys()]

        return mapping

    def get_mapper(self,
                   user: 'User',
                   name: str) -> 'PythonMapper':

        if self.mappings.get(user, None) is None:
            self.mappings[user] = {}

        if self.mappings[user].get(name) is None:
            self.mappings[user][name] = self.available_mappings[name].copy(user=user)

        return self.mappings[user][name].mapper

    def get_method_mapper(self,
                          user: 'User',
                          name: str) -> 'MethodMapper':

            if self.mappings.get(user, None) is None:
                self.mappings[user] = {}

            if self.mappings[user].get(name) is None:
                self.mappings[user][name] = self.available_mappings[name].copy(user=user)

            return self.mappings[user][name].method_mapper

    def get_view_manager(self, name: str) -> ViewManager:
        return self.available_mappings[name].view_manager

    def set_mapper(self,
                   user: 'User',
                   name: str,
                   mapper: 'PythonMapper') -> None:
        self.mappings[user][name].mapper = mapper
        mapper.method_mapper = self.mappings[user][name].method_mapper

    def add_package(self,
                    package: str,
                    user: 'User',
                    install: bool = True) -> None:
        try:
            new_package = importlib.import_module(package)
        except ImportError as e:
            ui.notify(f'Could not import package {package}')
            try:
                if install:
                    ui.notify(f'Installing package {package}')
                    user.logger.info(f'Installing package {package}')
                    import pip
                    pip.main(['install', package])
                    new_package = importlib.import_module(package)
            except Exception as e:
                user.logger.info(f'could not install package {package}: {e}')
                ui.notify(f'Could not install package {package}: {e}')
                return

        mapper = getattr(new_package, 'mapper', None)
        if mapper is None:
            ui.notify(f'Package {package} does not contain a mapper')
            user.logger.info(f'Package {package} does not contain a mapper')
            return

        method_mapper = getattr(new_package, 'method_mapper', None)
        if method_mapper is None:
            ui.notify(f'Package {package} does not contain a method_mapper')
            user.logger.info(f'Package {package} does not contain a method_mapper')
            return

        view_manager = getattr(new_package, 'view_manager', None)
        if view_manager is None:
            ui.notify(f'Package {package} does not contain a view_manager')
            user.logger.info(f'Package {package} does not contain a view_manager')
            return

        self.create_mapping(name=package,
                            mapper=getattr(new_package, 'mapper', None),
                            method_mapper=getattr(new_package, 'method_mapper', None),
                            view_manager=getattr(new_package, 'view_manager', None)
                            )

        if user.tool_select is not None:
            options = list(user.available_mappings.keys()) if user is not None else []
            options.sort()
            user.tool_select.set_options(options)

        user.logger.info(f'Added package {package}')
        ui.notify(f'Added ToolBox {package}')


# class MapperSelectDialog(ui.dialog):
#
#     def __init__(self, user=None) -> None:
#         super().__init__()
#         self._user = None
#
#         with self, ui.card():
#             ui.label('Select Toolbox:')
#             self.select = ui.select(options=list(self.user.available_mappings.keys()) if self.user is not None else [],
#                                     value=self.user.selected_mapper if self.user is not None else None,).classes('w-full')
#
#             with ui.checkbox('Load undefined components',
#                              value=True) as self.load_undefined:
#                 ui.tooltip('Load components that are not defined in the mapping')
#
#             with ui.row():
#                 ui.button('OK', on_click=self.select_mapper)
#                 ui.button('Cancel', on_click=self.cancel)
#
#         self.user = user
#
#     @property
#     def user(self) -> 'User':
#         return self._user
#
#     @user.setter
#     def user(self, value: 'User') -> None:
#         self._user = value
#         if self.user is not None:
#             self.select.options = list(self.user.available_mappings.keys())
#             self.select.value = self.user.selected_mapper
#         else:
#             self.select.options = []
#             self.select.value = None
#
#     def select_mapper(self, *args, **kwargs):
#
#         mapper = self.user.mapper_manager.get_mapping(self.user, self.select.value)
#         mapper.load_undefined = self.load_undefined.value
#         self.user.selected_mapper = self.select.value
#         self.user.project_manager.refresh_all_items()
#         ui.notify(f'Selected mapper: {self.user.selected_mapper}')
#         self.close()
#
#     def cancel(self, *args, **kwargs):
#         self.close()
#
#     def open(self, *args, **kwargs):
#         super().open(*args, **kwargs)


class MapperDropdown(ui.select):

    def __init__(
            self,
            user: 'User',
            *args,
            **kwargs,
    ) -> None:
        self._user = None

        # options = list(self.user.available_mappings.keys()) if self.user is not None else [],
        # value = self.user.selected_mapper if self.user is not None else None

        super().__init__(options=[],
                         on_change=self.mapper_changed,
                         *args,
                         **kwargs)

        self.user = user

    @property
    def user(self) -> 'User':
        return self._user

    @user.setter
    def user(self, value: 'User') -> None:
        self._user = value
        options = list(self.user.available_mappings.keys()) if self.user is not None else []
        options.sort()
        self.set_options(options)
        self.value = self.user.selected_mapper if self.user is not None else None

    def mapper_changed(self, *args, **kwargs):
        try:
            self.user.selected_mapper = self.value
            self.user.project_manager.refresh_all_items()
            ui.notify(f'Selected Toolbox: {self.user.selected_mapper}')
        except Exception as e:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logger.error(f'Error: {e}:\n'
                         f'{error}')
            ui.notify(f'Error: {e}:\n'
                      f'{error}')


class AddPackageDialog(ui.dialog):

    def __init__(self, user=None) -> None:
        super().__init__()
        self._user = None
        self.package_input = None
        self.upload_card = None

        with self, ui.card():
            ui.label('Add Toolbox')
            self.package_input = ui.input(placeholder='Package name').classes('w-full')
            self.install_cb = ui.checkbox('Install package', value=True)

            ui.separator()

            self.upload_card = ui.upload(label='Upload and install package',
                                         on_upload=self.upload_package,
                                         auto_upload=True).on(
                'finish', lambda: ui.notify('Finish!')
            ).classes('max-w-full')

            with ui.row():
                ui.button('OK', on_click=self.ok)
                ui.button('Cancel', on_click=self.cancel)

        self.user = user

    @property
    def user(self) -> 'User':
        return self._user

    @user.setter
    def user(self, value: 'User') -> None:
        self._user = value

    def ok(self, *args, **kwargs):
        package_name = self.package_input.value
        self.user.mapper_manager.add_package(package_name,
                                             self.user,
                                             install=self.install_cb.value)
        self.close()

    def cancel(self, *args, **kwargs):
        self.close()

    def open(self, *args, **kwargs):
        super().open(*args, **kwargs)

    def upload_package(self,
                       e: events.UploadEventArguments,
                       *args,
                       **kwargs):

        try:
            import zipfile
            def get_package_name_from_wheel(wheel_path):
                with zipfile.ZipFile(wheel_path, 'r') as zip_ref:
                    # Extract the METADATA file; it's typically located in the *.dist-info directory
                    dist_info = [f for f in zip_ref.namelist() if f.endswith('METADATA')]
                    if not dist_info:
                        return "No METADATA file found in the wheel"

                    # Read the METADATA file to extract the package name
                    metadata_file = dist_info[0]
                    with zip_ref.open(metadata_file) as metadata:
                        for line in io.TextIOWrapper(metadata):
                            if line.startswith('Name:'):
                                return line.split(':', 1)[1].strip()
                return "Package name not found in METADATA"

            shutil.copyfileobj(e.content, open(f'{project_dir}/{e.name}', 'wb'))
            ui.notify(f'Package {e.name} uploaded!')
            import pip
            try:
                pip.main(['install', f'{project_dir}/{e.name}'])
            except Exception as e:
                ui.notify(f'Could not install package {e.name}: {e}')
                return

            ui.notify(f'Package {e.name} installed!')

            self.upload_card.delete()
            ui.label(f'Package {e.name} installed!').classes('text-green-500')

            self.package_input.value = get_package_name_from_wheel(f'{project_dir}/{e.name}')
            self.package_input.update()
        except Exception as e:
            self.user.logger.error(f'Error uploading and installing package: {e}')
            ui.notify(f'Error uploading and installing package: {e}')

        # self.ok()



from nicegui import ui, app

from numpy import ndarray

from .. import user_manager
from ..core.user import User
from .. import core
from PySimultan2.simultan_object import SimultanObject
from PySimultan2.files import FileInfo
from PySimultan2.multi_values import simultan_multi_value_field_3d_to_numpy
from ..config import logger
from math import inf

from SIMULTAN.Data.MultiValues import (SimMultiValueField3D, SimMultiValueField3DParameterSource, SimMultiValueBigTable,
                                       SimMultiValueBigTableHeader, SimMultiValueBigTableParameterSource)



def get_value_content_type(value):
    if value is None:
        return 'None'
    if isinstance(value, str):
        return 'str'
    elif isinstance(value, int):
        return 'int'
    elif isinstance(value, float):
        return 'float'
    elif isinstance(value, SimultanObject):
        return 'Component'
    else:
        return 'None'


class ParameterEditDialog(object):

    def __init__(self, *args, **kwargs):
        self.sim_parameter = kwargs.get('sim_parameter', None)


class IntEditDialog(ParameterEditDialog):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.value_input = None
        self.min_input = None
        self.max_input = None
        self.unit_input = None

    def ui_content(self):
        with ui.row():
            self.value_input = ui.input(value=self.sim_parameter.Value if self.sim_parameter is not None else 0,
                                        validation=self.validate, label='Value')
            self.min_input = ui.input(value=self.sim_parameter.ValueMin if self.sim_parameter is not None else -inf,
                                      validation=self.validate, label='Min. Value')
            self.max_input = ui.input(value=self.sim_parameter.ValueMax if self.sim_parameter is not None else inf,
                                      validation=self.validate, label='Max. Value')
            self.unit_input = ui.input(value=self.sim_parameter.Unit if self.sim_parameter is not None else '',
                                       label='Unit')

    @staticmethod
    def validate(value):
        try:
            int(value)
        except ValueError:
            return "Value must be an integer!"

    @property
    def value(self):
        return int(self.value_input.value)

    @property
    def min(self):
        return int(self.min_input.value) if self.min_input.value != '-inf' else -inf

    @property
    def max(self):
        return int(self.max_input.value) if self.max_input.value != 'inf' else inf

    @property
    def unit(self):
        return self.unit_input.value


class FloatEditDialog(IntEditDialog):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def validate(value):
        try:
            float(value)
        except ValueError:
            return "Value must be an float!"

    @property
    def value(self):
        return float(self.value_input.value)

    @property
    def min(self):
        return float(self.min_input.value)

    @property
    def max(self):
        return float(self.max_input.value)


class StrEditDialog(ParameterEditDialog):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_input = None

    def ui_content(self):
        with ui.row():
            self.value_input = ui.input(value=self.sim_parameter.Value if self.sim_parameter is not None else '')

    @property
    def value(self):
        return self.value_input.value


class ComponentEditDialog(object):

    def __init__(self, *args, **kwargs):
        self.select_multiple = kwargs.get('select_multiple', True)
        self.component = kwargs.get('component', None)
        self.content = kwargs.get('content', None)
        self.parent = kwargs.get('parent', None)
        self.dialog = None

        self.class_select = None
        self.component_select = None

        self.component_name_map: dict = {}

    @property
    def value(self):
        if len(self.component_select.value) == 0:
            return None
        elif __len__ := len(self.component_select.value) == 1:
            return self.component_name_map[self.component_select.value[0]]
        else:
            return [self.component_name_map[x] for x in self.component_select.value]

    @ui.refreshable
    def ui_content(self):
        with ui.card():
            ui.label('Select component')
            mapper = user_manager[app.storage.user['username']].mapper
            classes = mapper.mapped_classes

            if isinstance(self.component, SimultanObject):
                select_value = self.component._taxonomy
            else:
                select_value = 'All'

            self.class_select = ui.select(['All', *classes.keys()],
                                          label='Select class',
                                          value=[select_value],
                                          multiple=self.select_multiple,
                                          on_change=self.on_cls_change,
                                          with_input=True
                                          ).classes('w-full').props('use-chips')
            self.component_ui_content()

    @ui.refreshable
    def component_ui_content(self):

        components = []
        self.component_name_map = {}

        mapper = user_manager[app.storage.user['username']].mapper

        if 'All' in self.class_select.value:
            for cls in mapper.mapped_classes.values():
                components.extend(cls.cls_instances)
        else:
            if isinstance(self.class_select.value, str):
                selected_val = [self.class_select.value]
            else:
                selected_val = self.class_select.value

            for cls in [mapper.get_mapped_class(x) for x in selected_val]:
                components.extend(cls.cls_instances)

        for component in components:
            self.component_name_map[f'{component.name} ({component.id})'] = component

        if isinstance(self.component, SimultanObject):
            select_value = [f'{self.component.name} ({self.component.id})']
        else:
            select_value = []

        self.component_select = ui.select(list(self.component_name_map.keys()),
                                          label='Select component',
                                          value=select_value,
                                          multiple=True,
                                          with_input=True,
                                          on_change=self.on_component_change
                                          ).classes('w-64').props('use-chips')

    def on_cls_change(self, event):
        self.component_ui_content.refresh()

    def on_component_change(self, event):
        pass


class AssetEditDialog(object):

    def __init__(self, *args, **kwargs):

        self.asset_select = None

        asset = kwargs.get('asset', None)

        self.asset = asset if isinstance(asset, FileInfo) else None
        self.content = kwargs.get('content', None)
        self.parent = kwargs.get('parent', None)
        self.dialog = None

        self.asset_name_map = {}

    def ui_content(self):
        with ui.card():
            ui.label('Select asset')
            assets = user_manager[app.storage.user['username']].asset_manager.items
            self.asset_name_map = {x.name: x for x in assets}
            self.asset_select = ui.select([x.name for x in assets],
                                          label='Select asset',
                                          value=self.asset.name if self.asset is not None else None,
                                          on_change=self.asset_select,
                                          with_input=True
                                          ).classes('w-64').props('use-chips')

    @property
    def value(self):
        if len(self.asset_select.value) == 0:
            return None

        return self.asset_name_map[self.asset_select.value]


class ArrayEditDialog(object):

    def __init__(self, *args, **kwargs):
        self.array_select = None
        asset = kwargs.get('array', None)

        self.array: SimMultiValueField3D = asset if isinstance(asset, ndarray) else None
        self.content = kwargs.get('content', None)
        self.parent = kwargs.get('parent', None)
        self.dialog = None

        self.array_name_map = {}

    @property
    def user(self) -> User:
        return user_manager[app.storage.user['username']]

    def ui_content(self):
        with ui.card():
            ui.label('Select Array')
            arrays = self.user.array_manager.np_items
            self.array_name_map = {x.Name: x for x in arrays.values()}
            self.array_select = ui.select([x.Name for x in arrays.values()],
                                          label='Select asset',
                                          value=self.array.Name if self.array is not None else None,
                                          on_change=self.array_select,
                                          with_input=True
                                          ).classes('w-64').props('use-chips')

    @property
    def value(self):
        if len(self.array_select.value) == 0:
            return None

        return self.array_name_map[self.array_select.value]


default_options = ['None', 'str', 'int', 'float', 'bool', 'Component', 'Asset', 'Array', 'Table']


class ContentTypeChooser(object):

    def __init__(self, *args, **kwargs):
        self.content_type = kwargs.get('content_type', str)
        self.select = None
        self.value = None
        self.options = kwargs.get('options', default_options)

        self.on_change = kwargs.get('on_change', None)

    def ui_content(self):
        self.select = ui.select(self.options,
                                value=self.content_type,
                                label='Content type',
                                on_change=self.on_change).classes('w-full')

    def on_change(self, event):
        if self.on_change is not None:
            self.on_change(event, self.value)


class ContentEditDialog(object):

    def __init__(self, *args, **kwargs):

        self._options = None

        self.select_multiple = kwargs.get('select_multiple', True)
        self.component = kwargs.get('component', None)
        self.raw_val = kwargs.get('raw_val', None)
        self.parent = kwargs.get('parent', None)
        self.content = kwargs.get('content', None)
        self.taxonomy = kwargs.get('taxonomy', None)
        self.object_mapper = kwargs.get('object_mapper', None)
        self.dialog = None

        if self.component is not None:
            val_type = get_value_content_type(self.component)
        else:
            val_type = 'None'

        self.content_type = ContentTypeChooser(on_change=self.on_type_change,
                                               content_type=val_type,
                                               options=self.options)

        self.options = kwargs.get('options', default_options)

        self.edit_dialog = None

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, value):
        self._options = value
        if self.content_type is not None:
            self.content_type.options = value

    def create_edit_dialog(self):
        with ui.dialog() as self.dialog, ui.card():
            # ui.label(f'Edit {self.content.name}')
            self.content_type.ui_content()
            self.ui_content()
            with ui.row():
                ui.button('Save', on_click=self.save)
                ui.button('Cancel', on_click=self.close)
            self.dialog.open()

    @ui.refreshable
    def ui_content(self):
        if self.content_type.select.value == 'str':
            self.edit_dialog = StrEditDialog(sim_parameter=self.raw_val)
            self.edit_dialog.ui_content()
        elif self.content_type.select.value == 'int':
            self.edit_dialog = IntEditDialog(sim_parameter=self.raw_val)
            self.edit_dialog.ui_content()
        elif self.content_type.select.value == 'float':
            self.edit_dialog = FloatEditDialog(sim_parameter=self.raw_val)
            self.edit_dialog.ui_content()
        elif self.content_type.select.value == 'Component':
            self.edit_dialog = ComponentEditDialog(component=self.component,
                                                   content=self.content,
                                                   parent=self.parent,
                                                   select_multiple=self.select_multiple)
            self.edit_dialog.ui_content()
        elif self.content_type.select.value == 'Asset':
            self.edit_dialog = AssetEditDialog(asset=self.component,
                                               content=self.content,
                                               parent=self.parent)
            self.edit_dialog.ui_content()
        elif self.content_type.select.value == 'Array':
            self.edit_dialog = ArrayEditDialog(array=self.component,
                                               content=self.content,
                                               parent=self.parent)
            self.edit_dialog.ui_content()
        else:
            ui.label('Not implemented content type!')

    def on_type_change(self):
        self.ui_content.refresh()

    def close(self, *args, **kwargs):
        if self.dialog is not None:
            self.dialog.close()
            self.dialog = None

    def save(self, *args, **kwargs):

        setattr(self.parent.component, self.content.property_name, self.edit_dialog.value)
        if self.content_type.select.value in ['int', 'float']:
            if self.raw_val is None:
                self.raw_val = self.parent.component.get_raw_attr(self.content.property_name)
            setattr(self.raw_val, 'ValueMin', self.edit_dialog.min)
            setattr(self.raw_val, 'ValueMax', self.edit_dialog.max)
            setattr(self.raw_val, 'Unit', self.edit_dialog.unit)

        logger.info(f'Updated {self.content.name} to {self.edit_dialog.value}')

        if self.parent is not None:
            self.parent.refresh()
        self.close()

        from ..views.detail_views import show_detail
        show_detail(value=self.parent.component)


class DictEditDialog(object):

    def __init__(self, *args, **kwargs):
        self.component = kwargs.get('component', None)
        self._key = None

        self.parent = kwargs.get('parent', None)
        self.dialog = None

        self.key_edit = None
        self.edit_dialog = None
        self.key = kwargs.get('key', None)

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value
        if self.key_edit is not None:
            self.key_edit.value = value

    def create_edit_dialog(self):
        with ui.dialog() as self.dialog, ui.card():
            ui.label(f'Edit {self.key}')
            self.ui_content()
            with ui.row():
                ui.button('Save', on_click=self.save)
                ui.button('Cancel', on_click=self.close)
            self.dialog.open()

    @ui.refreshable
    def ui_content(self):
        self.key_edit = ui.input(value=self.key, label='Key', on_change=self.on_key_change)
        self.edit_dialog = ContentEditDialog(component=getattr(self.parent.component, self.key) if self.key is not None else None,
                                             content=None,
                                             parent=self.parent,
                                             select_multiple=False)
        self.edit_dialog.content_type.ui_content()
        self.edit_dialog.ui_content()

    def on_key_change(self, event):
        self.key = self.key_edit.value

    def close(self, *args, **kwargs):
        if self.dialog is not None:
            self.dialog.close()
            self.dialog = None

    def save(self, *args, **kwargs):
        setattr(self.parent.component, self.key, self.edit_dialog.value)

        logger.info(f'Updated {self.key} to {self.edit_dialog.value}')

        if self.parent is not None:
            self.parent.ui_content.refresh()
        self.close()



from PySimultan2 import Content

contents = {}

contents['param_a'] = Content(text_or_key='param_a',
                              property_name='param_a',
                              type=None,
                              unit=None,
                              documentation='param_a')

contents['param_b'] = Content(text_or_key='param_b',
                              property_name='param_b',
                              type=None,
                              unit=None,
                              documentation='param_b')

contents['param_c'] = Content(text_or_key='param_c',
                              property_name='param_c',
                              type=None,
                              unit=None,
                              documentation='param_c')



from nicegui import ui

# Define columns and rows
columns = [{'name': 'item_id',
            'label': 'Item ID',
            'field': 'item_id',
            'align': 'left',
            'sortable': True},
           {'name': 'name',
            'label': 'Name',
            'field': 'name',
            'sortable': True},
           {'name': 'component_id',
            'label': 'Component ID',
            'field': 'component_id',
            'sortable': True},
           {'name': 'actions',
            'label': 'actions',
            'field': 'actions',
            'sortable': False}
           ]

rows = [{'item_id': '0', 'name': 'test1', 'component_id': '1894'},
        {'item_id': '1', 'name': 'test1', 'component_id': '1894'},
        {'item_id': '2', 'name': 'test1', 'component_id': '1894'},]

# Create the table
table = ui.table(columns=columns, rows=rows, row_key='name').classes('w-72')


table.add_slot('header', r'''
    <q-tr :props="props">
        <q-th auto-width />
        <q-th v-for="col in props.cols" :key="col.name" :props="props">
            {{ col.label }}
        </q-th>
    </q-tr>
''')

# Modify the table body slot for delete functionality
table.add_slot('body', r'''
    <q-tr :props="props">
        <q-td v-for="col in props.cols" :key="col.name" :props="props">
            {{ col.value }}
        </q-td>
        <q-td auto-width>
            <q-btn size="sm" color="negative" round dense
                   @click="$parent.$emit('delete', props)"
                   icon="delete" />
            <q-btn size="sm" color="positive" round dense
                   @click="$parent.$emit('up', props)"
                   icon="keyboard_arrow_up" />
            <q-btn size="sm" color="positive" round dense
                   @click="$parent.$emit('down', props)"
                   icon="keyboard_arrow_down" />
        </q-td>
    </q-tr>
''')


# Define and connect the event handler for delete
def handle_delete(props):
    print(f"Delete row: {props}")
    # Further deletion logic goes here


def handle_up(props):
    print(f"up row: {props}")


table.on('delete', handle_delete)
table.on('up', handle_up)
table.on('down', handle_up)

# Run the UI
ui.run()



from PySimultan2.geometry import GeometryModel
import os
import numpy as np
import struct


def rgba_to_hex(rgba):
    """
    Convert an RGBA color (with each component in 0-255 range) to a hexadecimal color string.

    Parameters:
        rgba (tuple): A tuple of four elements (R, G, B, A) where R, G, B are integers from 0 to 255
                      representing the color and A is a float from 0.0 to 1.0 representing the opacity.

    Returns:
        str: The hexadecimal color string including alpha channel as #RRGGBBAA.
    """
    r, g, b, a = rgba
    return f'#{r:02X}{g:02X}{b:02X}'


def create_stl_file(geometry_model: GeometryModel) -> dict[int: str]:
    stl_dir = os.path.join('/static/stl', str(geometry_model.key))
    if not os.path.exists(stl_dir):
        os.makedirs(stl_dir)

    file_lookup = {}

    for face in geometry_model.faces:
        filename = os.path.join(stl_dir, f"{face.id}.stl")
        file_lookup[face.id] = (os.path.join('/stl', str(geometry_model.key), f"{face.id}.stl"),
                                rgba_to_hex((getattr(face._wrapped_object.Color.Color,
                                                     key) for key in ('R', 'G', 'B', 'A')))
                                )
        vertices, triangles = face.triangulate()
        write_stl_from_numpy(vertices, triangles, filename=filename)

    return file_lookup


def write_stl_from_numpy(vertices, triangles, filename="output.stl"):
    """
    Write a binary STL file from numpy arrays of vertices and triangles.

    Parameters:
        vertices (np.array): Nx3 numpy array of vertices.
        triangles (np.array): Mx3 numpy array of triangle indices.
        filename (str): The filename to save the STL file to.
    """
    # Open the file in binary write mode
    with open(filename, "wb") as f:
        # Write the header
        f.write(b'\0' * 80)  # STL file header 80 bytes long
        f.write(struct.pack('<I', len(triangles)))  # Number of triangles

        # Write each triangle
        for tri in triangles:
            # Calculate the normal for the triangle
            v1 = vertices[tri[0]]
            v2 = vertices[tri[1]]
            v3 = vertices[tri[2]]
            normal = np.cross(v2 - v1, v3 - v1)
            normal = normal / np.linalg.norm(normal)

            # Pack the data in the right format and write to file
            f.write(struct.pack('<3f', *normal))  # Normal vector
            f.write(struct.pack('<3f', *v1))  # Vertex 1
            f.write(struct.pack('<3f', *v2))  # Vertex 2
            f.write(struct.pack('<3f', *v3))  # Vertex 3
            f.write(b'\0\0')  # Attribute byte count - not used

    print(f"File saved as {filename}")



import sys
import traceback
from nicegui import events, ui, app

from typing import Optional
from .. import user_manager
from .detail_views import show_detail

from SIMULTAN.Data import SimId
from PySimultan2.simultan_object import SimultanObject
from System import Guid


class GridView(object):

    columns = [
        {'headerName': 'Name',
         'field': 'name',
         'editable': True,
         'sortable': True,
         'filter': 'agTextColumnFilter',
         'floatingFilter': True,
         'checkboxSelection': True},
        {'headerName': 'ID',
         'field': 'id',
         'editable': False,
         'sortable': True,
         'filter': 'agTextColumnFilter',
         'floatingFilter': True},
        {'headerName': 'Mapped Class',
         'field': 'taxonomy',
         'editable': False,
         'sortable': True,
         'filter': 'agTextColumnFilter',
         'floatingFilter': True},
    ]

    def __init__(self, *args, **kwargs):

        self._data_model = None
        self.table: Optional[ui.aggrid] = None

        self.create_type_select: Optional[ui.select] = None
        self.create_type: Optional[str] = None             # type of object to create

    @property
    async def selected_instances(self):
        rows = await self.table.get_selected_rows()
        return [
            self.mapper.get_mapped_class(row['taxonomy'])._cls_instances_dict.get(SimId(Guid(row['id'].split('_')[0]),
                                                                                        int(row['id'].split('_')[1])),
                                                                                  None) for row in rows]

    @property
    def data_model(self):
        return self._data_model

    @data_model.setter
    def data_model(self, value):
        self._data_model = value
        self.ui_content.refresh()

    @property
    def mapper(self):
        return self.user.mapper

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @property
    def items(self) -> list[dict[str: str,
                                 str: str,
                                 str: str]]:
        items = []

        for taxonomy, cls in self.mapper.mapped_classes.items():
            if hasattr(cls, 'cls_instances'):
                for instance in cls.cls_instances:
                    items.append({'name': instance.name,
                                  'id': f'{instance.id.GlobalId}_{instance.id.LocalId}',
                                  'taxonomy': taxonomy})

        return items

    @ui.refreshable
    def ui_content(self):

        self.table = ui.aggrid({
            'columnDefs': self.columns,
            'rowData': self.items,
            'auto_size_columns': True,
            'rowSelection': 'multiple',
            'pagination': True,
            'paginationPageSize': 25,
            'paginationPageSizeSelector': [10, 25, 50, 100, 150, 200],
            'enableFilter': True,
        },
            auto_size_columns=True).on('cellClicked', lambda e: self.show_detail(e))

        # self.table.add_slot('body-cell-actions', r'''
        #                                                         <q-td key="actions" :props="props">
        #                                                             <q-btn size="sm" color="blue" round dense
        #                                                                 @click="$parent.$emit('show_detail', props)"
        #                                                                 icon="launch" />
        #                                                             <q-btn size="sm" color="blue" round dense
        #                                                                 @click="$parent.$emit('create_copy', props)"
        #                                                                 icon="file_copy" />
        #                                                         </q-td>
        #                                                     ''')
        #
        # self.table.on('show_detail', lambda e: self.show_detail(e))
        # self.table.on('create_copy', lambda e: self.create_copy(e))

        self.table.classes('w-full h-full')

        with ui.row().classes('w-full gap-0'):
            with ui.column().classes('w-1/3 gap-0'):
                self.ui_create_new_content()
            with ui.column().classes('w-1/2 gap-0'):
                self.user.method_mapper.ui_content()

            ui.space()

            with ui.button(on_click=self.user.project_manager.refresh_all_items,
                           icon='update').props('fab color=accent').classes('h-3 justify=end items=center'):
                ui.tooltip('Update data')

    def ui_create_new_content(self):

        with ui.expansion(icon='add',
                          text='Create new').classes('w-full') as exp:

            with ui.column().classes('w-full'):

                if not self.user.mapper.mapped_classes:
                    _ = [self.user.mapper.get_mapped_class(x) for x in self.user.mapper.registered_classes.keys()]

                options = [taxonomy for taxonomy in self.user.mapper.mapped_classes.keys()
                           if taxonomy not in self.user.mapper.undefined_registered_classes.keys()]

                self.create_type_select = ui.select(options=options,
                                                    value=options[0]
                                                    if self.user.mapper.mapped_classes.keys() else None,
                                                    with_input=True,
                                                    label='Mapped Class').bind_value(self,
                                                                                     'create_type'
                                                                                     ).classes('w-full')
                with self.create_type_select.add_slot('append'):
                    with ui.button(icon='play_circle', on_click=self.create_new_item
                                   ).classes('h-5 justify=end items=center'):
                        ui.tooltip('Create')

    def create_new_item(self, e):

        try:

            if None in (self.data_model, self.create_type):
                ui.notify('No data model selected! Please load a data model first.')
                return

            cls = self.user.mapper.get_mapped_class(self.create_type)
            if cls is None:
                ui.notify(f'Could not find class for taxonomy {self.create_type}.')
                return
            new_item: SimultanObject = cls(data_model=self.data_model,
                                           object_mapper=self.mapper,
                                           )
            new_item.name = f'New {cls.__name__}_{new_item.id}'

            self.user.logger.info(f'Created new {cls.__name__}, ID {new_item.id}.')

            self.user.project_manager.mapped_data.append(new_item)

            self.data_model.save()
            self.add_item_to_view(new_item)
            self.mapper.get_mapped_class(self.create_type)._cls_instances_dict.get(new_item.id, None)

            show_detail(value=new_item)
            ui.notify(f'Created new {cls.__name__}, ID {new_item.id}.')
        except Exception as e:
            err = '\n'.join(traceback.format_exception(*sys.exc_info()))
            ui.notify(f'Error creating new item: {e}')
            self.user.logger.error(f'Error creating new item: {e}:\n{err}')

    def selected_rows(self):
        rows = self.table.get_selected_rows()
        if rows:
            for row in rows:
                ui.notify(f"{row['name']}, {row['age']}")
        else:
            ui.notify('No rows selected.')

    def show_detail(self, e):

        component_id = e.args['data']['id']
        sim_id = SimId(Guid(component_id.split('_')[0]), int(component_id.split('_')[1]))
        instance = self.mapper.get_mapped_class(e.args['data']['taxonomy'])._cls_instances_dict.get(sim_id, None)
        if instance is None:
            return
        show_detail(value=instance)

    def add_item_to_view(self, component: SimultanObject):
        self.ui_content.refresh()

#
#
#
# def rename(e: events.GenericEventArguments) -> None:
#     for row in rows:
#         if row["id"] == e.args["id"]:
#             row.update(e.args)
#     ui.notify(f"Table.rows is now: {table.rows}")
#
#
# def delete(e: events.GenericEventArguments) -> None:
#     rows[:] = [row for row in rows if row["id"] != e.args["id"]]
#     ui.notify(f"Delete {e.args['id']}")
#     table.update()
#
#
# def addrow() -> None:
#     newid = max(dx["id"] for dx in rows) + 1
#     rows.append({"id": newid, "name": "New guy", "age": 21})
#     ui.notify(f"Added new row with id {newid}")
#     table.update()
#
#
# table = ui.table(columns=columns, rows=rows, row_key="name").classes("w-72")
# table.add_slot(
#     "header",
#     r"""
#     <q-tr :props="props">
#         <q-th auto-width />
#         <q-th v-for="col in props.cols" :key="col.name" :props="props">
#             {{ col.label }}
#         </q-th>
#     </q-tr>
# """,
# )
# table.add_slot(
#     "body",
#     r"""
#     <q-tr :props="props">
#         <q-td auto-width >
#             <q-btn size="sm" color="warning" round dense icon="delete" :props="props"
#                 @click="() => $parent.$emit('delete', props.row)" >
#         </q-td>
#         <q-td key="name" :props="props">
#             {{ props.row.name }}
#             <q-popup-edit v-model="props.row.name" v-slot="scope"
#                 @update:model-value="() => $parent.$emit('rename', props.row)" >
#                 <q-input v-model="scope.value" dense autofocus counter @keyup.enter="scope.set" />
#             </q-popup-edit>
#         </q-td>
#         <q-td key="age" :props="props" class="w-8">
#             {{ props.row.age }}
#             <q-popup-edit v-model="props.row.age" v-slot="scope"
#                 @update:model-value="() => $parent.$emit('rename', props.row)" >
#                 <q-input v-model.number="scope.value" type="number" dense autofocus counter @keyup.enter="scope.set" />
#             </q-popup-edit>
#         </q-td>
#     </q-tr>
#     """,
# )
# table.add_slot(
#     "bottom-row",
#     r"""
#     <q-tr :props="props">
#         <q-td colspan="3" class="text-center">
#             <q-btn color="accent" icon="add" class="w-full" @click="() => $parent.$emit('addrow')"/>
#         </q-td>
#     </q-tr>
#     """,
# )
# table.on("rename", rename)
# table.on("delete", delete)
# table.on("addrow", addrow)
#
# ui.run()



from nicegui import Client, app, ui, events
from .type_view import TypeView

from ..core.edit_dialog import ContentEditDialog


class ParameterView(TypeView):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.taxonomy = kwargs.get('taxonomy', None)
        self.raw_val = kwargs.get('raw_val', None)
        self.content = kwargs.get('content', None)

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, value):
        self._component = value

    @ui.refreshable
    def ui_content(self):

        with ui.item().classes('w-full h-full'):
            with ui.item_section():
                ui.label(f'{self.component}')

            if self.raw_val is None:
                with ui.item_section():
                    ui.label(f': {None}')
            else:
                if isinstance(self.raw_val.Value, (int, float)):
                    with ui.item_section():
                        if hasattr(self.raw_val, 'Unit'):
                            ui.label(f': {self.raw_val.Unit}')
                        else:
                            ui.label('')
                    with ui.item_section():
                        if hasattr(self.raw_val, 'ValueMin'):
                            ui.label(f': {self.raw_val.ValueMin}')
                        else:
                            ui.label('')
                    with ui.item_section():
                        if hasattr(self.raw_val, 'ValueMax'):
                            ui.label(f': {self.raw_val.ValueMax}')
                        else:
                            ui.label('')

            with ui.item_section():
                ui.button(icon='edit', on_click=self.edit).classes('q-ml-auto')

    def edit(self, event):
        edit_dialog = ContentEditDialog(component=self.component,
                                        raw_val=self.raw_val,
                                        parent=self.parent,
                                        content=self.content,
                                        taxonomy=self.taxonomy)
        edit_dialog.create_edit_dialog()



import os

from PySimultanUI.src.pysimultanui import run_ui, user_manager

from toolbox_1.mapping import mapper as mapper1, method_mapper as method_mapper1, view_manager as view_manager1
from toolbox_2.mapping import mapper as mapper2, method_mapper as method_mapper2, view_manager as view_manager2

project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)


mapping = user_manager.mapper_manager.create_mapping(name='Toolbox 1',
                                                     mapper=mapper1,
                                                     method_mapper=method_mapper1,
                                                     view_manager=view_manager1)

mapping2 = user_manager.mapper_manager.create_mapping(name='Toolbox 2',
                                                      mapper=mapper2,
                                                      method_mapper=method_mapper2,
                                                      view_manager=view_manager2)

run_ui()
print('Done!')



import FreeCAD
import Part as FCPart

from typing import List, Optional, Union

from .scene import ExtendedScene
from .three_views import display_vertices, display_edges, display_wire, display_face, display_solid
from nicegui.elements.scene_objects import Group


def display_feature(feature: FCPart.Feature,
                    scene: ExtendedScene
                    ) -> tuple[Group, tuple[float, float, float]]:

    x_height = feature.Shape.BoundBox.XMax
    y_height = feature.Shape.BoundBox.YMax
    z_height = feature.Shape.BoundBox.ZMax

    rgba_string = feature.ShapeMaterial.Properties['AmbientColor']
    rgba_values = rgba_string.strip('()').split(',')
    rgb = [int(float(value.strip()) * 255) for value in rgba_values[:3]]
    black_color = [255, 255, 255]

    with scene.group() as group:
        if hasattr(feature.Shape, 'Vertexes'):
            display_vertices(feature.Shape.Vertexes,
                             scene,
                             str(feature.ID),
                             colors=[black_color] * feature.Shape.Vertexes.__len__()
                             )

        if hasattr(feature.Shape, 'Edges'):
            display_edges(feature.Shape.Edges,
                          scene,
                          str(feature.ID),
                          black_color)

        if hasattr(feature.Shape, 'Wires'):
            for wire in feature.Shape.Wires:
                display_wire(wire,
                             scene,
                             str(feature.ID),
                             black_color)

        if hasattr(feature.Shape, 'Faces'):
            for face in feature.Shape.Faces:
                display_face(face, scene, str(feature.ID), rgb)

        if hasattr(feature.Shape, 'Solids'):
            for solid in feature.Shape.Solids:
                display_solid(solid, scene, str(feature.ID), rgb)

    scene.move_camera(x=x_height, y=-y_height, z=z_height * 2, duration=2)

    return group, (x_height, y_height, z_height)



import shutil
import tempfile
import os
from typing import Optional
from nicegui import ui, events, app
from PySimultan2.geometry import GeometryModel

from ... import user_manager
from . import GeometryView
from ..type_view_manager import TypeViewManager
from ... import user_manager


try:
    try:
        from ...core import freecad_utils
    except ImportError:
        pass

    import FreeCAD
    import Part as FCPart
    freecad_supported = True

    from ...core.freecad_utils.tools import import_freecad

except ImportError:
    freecad_supported = False


class GeometryManager(TypeViewManager):

    columns = [{'name': 'id',
                'label': 'Key',
                'field': 'id',
                'sortable': True,
                'align': 'left'},
               {'name': 'name',
                'label': 'Name',
                'field': 'name',
                'sortable': True,
                'align': 'left'},
               {'name': 'file_name',
                'label': 'File Name',
                'field': 'file_name',
                'sortable': True,
                'align': 'left'},
               {'name': 'file_size',
                'label': 'File Size',
                'field': 'file_size',
                'sortable': True,
                'align': 'left'},
               {'name': 'actions',
                'label': 'Actions',
                'field': 'actions',
                'sortable': False,
                'align': 'left'},
               ]

    item_view_name = 'Geometry Models'
    item_view_cls = GeometryView
    cls = GeometryModel

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_model_dialog = None
        self.table: Optional[ui.table] = None

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @property
    def mapper(self):
        return self.user.mapper

    @property
    def project_manager(self):
        return self.user.project_manager

    @property
    def selected_items(self):
        selected_keys = [int(x['id']) for x in self.table.selected]
        return [x for x in self.items if x.key in selected_keys]

    def update_items(self) -> list[any]:
        if self.data_model is None:
            return []

        return [GeometryModel(wrapped_object=x,
                              object_mapper=self.mapper,
                              data_model=self.data_model) for x in
                self.data_model.models.values()]

    def button_create_ui_content(self):
        ui.button('Create new', on_click=self.create_new_model)

    @ui.refreshable
    def ui_content(self):
        self.items = self.update_items()
        self.methods_ui_content()

        rows = [{'id': f'{item.key}',
                 'name': item.name,
                 'file_name': item.resource_file_entry.Name if item.resource_file_entry is not None else 'None',
                 'file_size': f'{os.path.getsize(item.resource_file_entry.File.FullPath) / 1024:.2f} kB'
                 }
                for item in self.items]

        with ui.table(columns=self.columns,
                      rows=rows,
                      title='Geometry Models',
                      selection='single',
                      row_key='id',
                      pagination={'rowsPerPage': 10, 'sortBy': 'id', 'page': 1}).classes('w-full bordered') as self.table:

            # self.table.add_slot('body', r'''
            #                     <q-tr :props="props">
            #                         <q-td v-for="col in props.cols" :key="col.name" :props="props">
            #                             {{ col.value }}
            #                         </q-td>
            #                         <q-td auto-width>
            #                             <q-btn size="sm" color="blue" round dense
            #                                    @click="$parent.$emit('show_detail', props)"
            #                                    icon="launch" />
            #                         </q-td>
            #                     </q-tr>
            #                 ''')

            self.table.add_slot('body-cell-actions', r'''
                                            <q-td key="actions" :props="props">
                                                <q-btn size="sm" color="blue" round dense
                                                    @click="$parent.$emit('show_detail', props)"
                                                    icon="launch" />
                                            </q-td>
                                        ''')

            self.table.on('show_detail', self.show_detail)

        self.button_create_ui_content()

    def show_detail(self, e: events.GenericEventArguments):
        from ..detail_views import show_detail
        geo_model: GeometryModel = next((x for x in self.items if x.key == int(e.args['row']['id'])), None)
        show_detail(geo_model)

    def add_item_to_view(self, item: GeometryModel):

        if item.key not in (int(x['id']) for x in self.table.rows):
            self.table.add_rows({'id': f'{item.key}',
                                 'name': item.name,
                                 'file_name': item.resource_file_entry.Name if item.resource_file_entry is not None else 'None',
                                 'file_size': f'{os.path.getsize(item.resource_file_entry.File.FullPath) / 1024:.2f} kB'
                                 })

    # @ui.refreshable
    # def ui_content(self):
    #
    #     with ui.row().classes('w-full h-full') as self.expansion:
    #
    #     # with ui.expansion(icon='format_list_bulleted',
    #     #                   text=f'{self.item_view_name if self.item_view_name is not None else self.cls._taxonomy} '
    #     #                        f'({len(self.items)})').classes('w-full h-full') as self.expansion:
    #     #
    #     #     self.expansion.bind_text_from(self,
    #     #                                   'items',
    #     #                                   lambda x: f'{self.item_view_name} ({len(x)})'
    #     #                                   )
    #
    #         self.methods_ui_content()
    #         self.items_ui_element = ui.row().classes('w-full h-full')
    #
    #         if len(self.items) == 0:
    #             with self.items_ui_element:
    #                 ui.label('No items to display')
    #         else:
    #             for item in self.items:
    #                 self.add_item_to_view(item)
    #
    #         with ui.row().classes('w-full h-full'):
    #             self.button_create_ui_content()
    #
    #     # super().ui_content()

    def create_new_model(self):
        def create_new_model_action():
            try:

                new_model = GeometryModel(data_model=self.user.data_model,
                                          name=self.new_model_dialog.project_name.value)
                self._geometry_models = None
                self._geometry_model_views = None
                self.new_model_dialog.close()
                ui.notify('New model created', type='positive')
                self.add_item_to_view(new_model)
                self.items = self.update_items()
            except Exception as e:
                ui.notify(f'Error creating new model: {e}', type='negative')

        with ui.dialog() as self.new_model_dialog, ui.card():
            ui.label('Create new fc_geometry model')
            self.new_model_dialog.project_name = ui.input('Name')
            with ui.row().classes('justify-between'):
                ui.button('Create', on_click=create_new_model_action)
                ui.button('Cancel', on_click=self.new_model_dialog.close)

        self.new_model_dialog.open()

    def import_from_file(self):
        # check if only one model is selected:
        if len(self.selected_items) != 1:
            ui.notify('Please select exactly one model to import to', type='negative')
            return

        # create dialog to upload file
        with ui.dialog() as dialog, ui.card():
            dialog.scale_input = ui.input(label='Scale Model', value='0.001').classes('w-full')

            upload = ui.upload(label='Upload asset',
                               on_upload=lambda e: self.upload_geometry(e, dialog=dialog)).on(
                'finish', lambda: ui.notify('Finish!')
            ).classes('max-w-full')
            upload.scale_input = dialog.scale_input
            ui.button('Cancel', on_click=lambda e: dialog.close()).classes('mt-4')
        dialog.open()

    def upload_geometry(self,
                        e: events.UploadEventArguments,
                        dialog: ui.dialog,
                        *args,
                        **kwargs):

        dialog.close()

        try:

            with tempfile.TemporaryDirectory() as tmpdirname:
                temp_file_path = os.path.join(tmpdirname, e.name)
                with open(temp_file_path, 'wb') as f:
                    shutil.copyfileobj(e.content, f)
                ui.notify(f'Project {e.name} uploaded!')

                if e.name.endswith('.fcstd') or e.name.endswith('.FCStd'):
                    doc = FreeCAD.open(temp_file_path)
                else:
                    doc = FreeCAD.newDocument()
                    s = FCPart.read(temp_file_path)
                    doc.addObject("Part::Feature", "Part").Shape = s

                num_created = import_freecad(doc=doc,
                                             geo_model=self.selected_items[0],
                                             data_model=self.data_model,
                                             object_mapper=self.user.mapper,
                                             scale=float(e.sender.scale_input.value))

            # local_path = f'/tmp/{e.name}'
            # shutil.copyfileobj(e.content, open(local_path, 'wb'))
            # ui.notify(f'Project {e.name} uploaded!')
            #
            # # open file in FreeCAD
            # doc = FreeCAD.open(local_path)
            # num_created = import_freecad(doc=doc,
            #                              geo_model=self.selected_items[0],
            #                              data_model=self.data_model,
            #                              object_mapper=self.user.mapper,
            #                              scale=float(e.sender.scale_input.value))

            ui.notify(f'Imported {num_created[0]} Vertices, '
                      f'{num_created[1]} Edges, '
                      f'{num_created[2]} EdgeLoops, '
                      f'{num_created[3]} Faces, '
                      f'{num_created[4]} Volumes.', type='positive')

            # create new fc_geometry model
            # self.ui_content.refresh()
        except Exception as e:
            ui.notify(f'Error uploading file: {e}', type='negative')

    def methods_ui_content(self):
        with ui.expansion(icon='code', text='Methods').classes('w-full bg-stone-300'):
            with ui.list().classes('w-full h-full'):
                if freecad_supported:
                    with ui.row().classes('w-full h-full'):
                        ui.label('Import from file')
                        ui.button('Import', on_click=self.import_from_file).classes('q-ml-auto')
                else:
                    ui.label('No Methods registered (FreeCAD not supported)')



from nicegui import ui
from logging import getLogger
from PySimultan2.default_types import ComponentDictionary
from ..type_view_manager import TypeViewManager


logger = getLogger('py_simultan_ui')


class ComponentDictManager(TypeViewManager):

    cls = ComponentDictionary

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def button_create_ui_content(self):
        ui.button('Create new Component dict', on_click=self.create_new_item, icon='add')

    def create_new_item(self, event):
        if self.data_model is None:
            ui.notify('No data model selected! Please load a data model first.')
            return

        new_item = self.cls(data_model=self.data_model)
        self.add_item_to_view(new_item)

    def update_items(self) -> list[any]:
        if self.cls is None:
            return []

        mapped_cls = self.mapped_cls

        if hasattr(mapped_cls, 'cls_instances'):
            logger.info(f'Found {len(mapped_cls.cls_instances)} instances for TVM {self.taxonomy}')
            return mapped_cls.cls_instances
        return []



import asyncio
import traceback
import sys
from typing import Any
from ..config import logger
from nicegui import ui, run, app, events
from .tools import ConfirmationDialog

from PySimultan2.simultan_object import SimultanObject


class UnknownClass(object):
    def __init__(self, *args, **kwargs):
        self.cls_name: type = kwargs.get('cls_name', 'Unknown class')
        self.cls: type = kwargs.get('cls', None)
        self.mapper = kwargs.get('mapper', None)


class UnmappedMethod(object):

    def __init__(self, *args, **kwargs):
        self.name: str = kwargs.get('name', 'Unnamed method')
        self.method = kwargs.get('method')
        self.args: list[Any] = kwargs.get('args', [])       # list with description of the arguments
        self.kwargs: dict[str:Any] = kwargs.get('kwargs', dict())       # dict with description of the keyword arguments

        self.add_data_model_to_kwargs = kwargs.get('add_data_model_to_kwargs', False)
        self.add_user_to_kwargs = kwargs.get('add_user_to_kwargs', False)

    @property
    def user(self):
        from .. import user_manager
        return user_manager[app.storage.user['username']]

    @property
    def logger(self):
        return self.user.logger

    @property
    def data_model(self):
        return self.user.data_model

    def ui_content(self):
        with ui.item().classes('w-full h-full'):
            with ui.item_section():
                ui.label(self.name)
            with ui.item_section():
                ui.button(on_click=self.run, icon='play_circle').classes('q-ml-auto')

    async def run(self, *args, **kwargs):

        self.logger.info(f'Running global method {self.name}')

        if self.data_model is None:
            self.logger.error('No project loaded!')
            ui.notify('No project loaded!', type='negative')
            return

        n = ui.notification(timeout=None)
        n.spinner = True
        n.message = f'Running method {self.name}'

        if self.add_data_model_to_kwargs:
            kwargs['data_model'] = self.data_model
        if self.add_user_to_kwargs:
            kwargs['user'] = self.user

        try:
            if self.kwargs.get('io_bound', True):
                await run.io_bound(self.method, *args, *self.args, **kwargs, **self.kwargs)
            else:
                self.method(*args, *self.args, **kwargs, **self.kwargs)
            n.type = 'positive'
            n.message = f'Successfully ran method {self.name}!'
            n.spinner = False
            await asyncio.sleep(1)
            n.dismiss()
            self.logger.info(f'Method {self.name} ran successfully!')
        except Exception as e:
            self.logger.error(f'Error running global method {self.name}:\n{e}'
                              f'{traceback.print_exception(*sys.exc_info())}')
            n.message = f'Error running method {self.name}: {e}'
            n.spinner = False
            await asyncio.sleep(2)
            n.dismiss()


class MappedMethod(object):

    def __init__(self,
                 method,
                 cls,
                 name=None,
                 args=None,
                 kwargs=None,
                 add_data_model_to_kwargs=False,
                 add_user_to_kwargs=False,
                 *s_args,
                 **s_kwargs):
        self.name: str = name if name is not None else 'Unnamed method'
        self.method = method
        self.cls: type = cls
        self.args: list[str] = args if args is not None else []             # list with description of the arguments
        self.kwargs: dict[str:str] = kwargs if kwargs is not None else {}       # dict with description of the keyword arguments
        self.add_data_model_to_kwargs = add_data_model_to_kwargs
        self.add_user_to_kwargs = add_user_to_kwargs

    @property
    def user(self):
        from .. import user_manager
        return user_manager[app.storage.user['username']]

    @property
    def data_model(self):
        return self.user.data_model

    @property
    def logger(self):
        return self.user.logger

    def ui_content(self):
        with ui.item():
            with ui.item_section():
                ui.label(self.name)
            with ui.item_section():
                ui.button(on_click=self.run, icon='play_circle').classes('q-ml-auto')

    async def run(self, *args, **kwargs):

        selected_instances: list[SimultanObject] = kwargs.get('selected_instances', None)
        kwargs.pop('selected_instances', None)

        self.logger.info(f'Running method {self.name} of {self.cls.__name__}')

        if selected_instances is None:
            self.logger.error('No instances selected!')
            ui.notify('No instances selected!')
            await asyncio.sleep(1)
            return

        if not all(instance._taxonomy == self.cls._taxonomy for instance in selected_instances):
            wrong_names = r' \n'.join([str((instance.name, str(instance.id))) for instance in selected_instances])
            self.logger.error(f'Not all selected instances are of the type {self.cls.__name__}!:'
                              f"{wrong_names}")

            ui.notify(f'Not all selected instances are of the type {self.cls.__name__}!:'
                      f"{wrong_names}",
                      type='negative',
                      close_button=True,
                      multi_line=True)
            await asyncio.sleep(1)
            return

        from .. import user_manager

        if user_manager[app.storage.user['username']].data_model is None:
            ui.notify('No project loaded!', type='negative')
            self.logger.error('No project loaded!')
            return

        if not selected_instances:
            ui.notify('No instances selected!')
            return

        def additional_content_fcn():
            ui.label('Selected instances:')
            for instance in selected_instances:
                with ui.row():
                    ui.label(f'{instance.name} ({instance.id})')

        if self.add_data_model_to_kwargs:
            kwargs['data_model'] = self.data_model
        if self.add_user_to_kwargs:
            kwargs['user'] = self.user

        result = await ConfirmationDialog(f'Are you sure you want to run {self.name}?',
                                          'Yes',
                                          'No',
                                          additional_content_fcn=additional_content_fcn)

        self.logger.info(f'User confirmed running method {self.name} of {self.cls.__name__}')

        if result == 'Yes':
            n = ui.notification(timeout=None)
            n.spinner = True
            n.message = f'Running method {self.name} of {self.cls.__name__}'
            logger.info(f'Running method {self.name} of {self.cls.__name__}')
            await asyncio.sleep(0.01)
            for instance in selected_instances:
                logger.info(f'Running method {self.name} on {instance.name} {instance.id}')

                try:
                    if kwargs.get('io_bound', True):
                        await run.io_bound(self.method, instance, *args, *self.args, **kwargs, **self.kwargs)
                    else:
                        self.method(*args, *self.args, instance, **kwargs, **self.kwargs)

                # self.method(instance, *self.args, **self.kwargs)
                except Exception as e:
                    err = '\n'.join(traceback.format_exception(*sys.exc_info()))

                    self.logger.error(f'Error running method {self.name} on {instance.name}: {e}'
                                      f'{err}')
                    ui.notify(f'Error running method {self.name} on {instance.name}: {e}', type='negative')
                    continue

                # getattr(instance, self.name)(*self.args, **self.kwargs)
                # self.method(instance, *self.args, **self.kwargs)
            self.logger.info(f'Method {self.name} run on {len(selected_instances)} instances!')
            ui.notify(f'Method {self.name} run on {len(selected_instances)} instances!')
            n.message = 'Done!'
            n.spinner = False
            await asyncio.sleep(1)
            # self.method(*args, **kwargs)
            n.dismiss()


class MethodMapper(object):

    def __init__(self, *args, **kwargs):

        self.mapped_methods = {}
        self.unmapped_methods = []
        self.mapper = kwargs.get('mapper', None)
        self.card = None

        self.global_methods_table = None
        self.mapped_methods_table = None

        self.cls_select = None

    @property
    def selected_cls(self):
        return next((cls for cls in self.mapped_methods.keys()
                     if not isinstance(cls, UnknownClass)
                     and cls.__name__ == self.cls_select.value), None)

    @property
    def user(self):
        from .. import user_manager
        return user_manager[app.storage.user['username']]

    def register_method(self,
                        method: callable,
                        name='unnamed_method',
                        args=(),
                        kwargs={},
                        cls: type = None,
                        add_data_model_to_kwargs=False,
                        add_user_to_kwargs=False,
                        *s_args,
                        **s_kargs):
        if cls is None:
            self.unmapped_methods.append(UnmappedMethod(name=name,
                                                        method=method,
                                                        args=args,
                                                        add_data_model_to_kwargs=add_data_model_to_kwargs,
                                                        add_user_to_kwargs=add_user_to_kwargs,
                                                        kwargs=kwargs,
                                                        *s_args,
                                                        **s_kargs),
                                         )
            return

        if cls not in self.mapped_methods:
            self.mapped_methods[cls] = []
        self.mapped_methods[cls].append(MappedMethod(cls=cls,
                                                     name=name,
                                                     method=method,
                                                     args=args,
                                                     kwargs=kwargs,
                                                     add_data_model_to_kwargs=add_data_model_to_kwargs,
                                                     add_user_to_kwargs=add_user_to_kwargs,
                                                     *s_args,
                                                     **s_kargs),
                                        )

    @ui.refreshable
    def ui_content(self):

        with ui.expansion(icon='play_circle',
                          text='Methods').classes('w-full'):

            with ui.expansion(icon='public',
                              text='Global Methods').classes('w-full') as self.card:

                columns = [{'name': 'id', 'label': '#', 'field': 'id', 'sortable': True},
                           {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                           {'name': 'actions', 'label': 'Actions', 'field': 'actions', 'sortable': False}]

                rows = [{'id': i, 'name': method.name} for i, method in enumerate(self.unmapped_methods)]

                self.global_methods_table = ui.table(columns=columns,
                                                     rows=rows,
                                                     pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1},
                                                     row_key='id').classes('w-full h-full bordered')

                self.global_methods_table.add_slot('body-cell-actions', r'''
                                                        <q-td key="actions" :props="props">
                                                            <q-btn size="sm" color="blue" round dense
                                                                @click="$parent.$emit('run_global_method', props)"
                                                                icon="play_circle" />
                                                        </q-td>
                                                    ''')
                self.global_methods_table.on('run_global_method', self.run_global_method)

                self.resolve_classes()

            with ui.expansion(icon='public',
                              text='Mapped Methods').classes('w-full'):

                options = [cls.__name__ for cls in self.mapped_methods.keys()
                           if not isinstance(cls, UnknownClass)]
                self.cls_select = ui.select(label='Select class',
                                            options=options,
                                            value=options[0] if options else None,
                                            on_change=self.ui_method_select_content.refresh)
                self.cls_select.classes('w-full')
                self.ui_method_select_content()

    async def run_method(self, e: events.GenericEventArguments):
        selected_instances = await self.user.grid_view.selected_instances
        await self.mapped_methods[self.selected_cls][int(e.args['key'])].run(selected_instances=selected_instances)

    @ui.refreshable
    def ui_method_select_content(self):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True}]

        if self.selected_cls is None:
            rows = []
        else:
            rows = [{'id': i,
                     'name': method.name}
                    for i, method in enumerate(self.mapped_methods[self.selected_cls])]

        methods_table = ui.table(columns=columns,
                                 rows=rows,
                                 row_key='id').classes('w-full bordered')

        methods_table.add_slot('body', r'''
                                    <q-tr :props="props">
                                        <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                            {{ col.value }}
                                        </q-td>
                                        <q-td auto-width>
                                            <q-btn size="sm" color="blue" round dense
                                                   @click="$parent.$emit('run_method', props)"
                                                   icon="play_circle" />
                                        </q-td>
                                    </q-tr>
                                ''')


        methods_table.on('run_method', self.run_method)

    async def run_global_method(self, e: events.GenericEventArguments):

        method = self.unmapped_methods[e.args['key']]
        await method.run()

    def resolve_classes(self):

        mapped_methods = self.mapped_methods.copy()

        for cls, methods in self.mapped_methods.items():
            if isinstance(cls, UnknownClass):
                for method in methods:
                    new_cls = self.mapper.mapped_classes.get(method.method.__qualname__.split('.')[-2], None)
                    if new_cls is None:
                        continue
                    method.cls = new_cls
                    if new_cls not in mapped_methods:
                        mapped_methods[new_cls] = []
                    mapped_methods[new_cls].append(method)
                    mapped_methods[cls].remove(method)

        self.mapped_methods = mapped_methods

    def copy(self):
        new_method_mapper = MethodMapper()
        new_method_mapper.mapped_methods = self.mapped_methods
        new_method_mapper.unmapped_methods = self.unmapped_methods
        new_method_mapper.mapper = self.mapper
        return new_method_mapper


method_mapper = MethodMapper()


def mapped_method(name=None, *args, **kwargs):

    dat_dict = {}
    dat_dict['name'] = name

    def wrapper(func):
        # vars(sys.modules[func.__module__])[func.__qualname__.split('.')
        method_mapper.register_method(method=func,
                                      name=dat_dict['name'],
                                      args=args,
                                      kwargs=kwargs,
                                      cls=UnknownClass(cls_name=func.__qualname__))
        return func
    return wrapper



from typing import Callable, Dict, Union
from logging import getLogger
from nicegui import background_tasks, helpers, ui

logger = getLogger('py_simultan_ui')


class RouterFrame(ui.element, component='router_frame.js'):
    pass


class Router():

    def __init__(self) -> None:
        self.routes: Dict[str, Callable] = {}
        self.content: ui.element = None

    def add(self, path: str):
        def decorator(func: Callable):
            self.routes[path] = func
            return func
        return decorator

    def open(self, target: Union[Callable, str]) -> None:
        if isinstance(target, str):
            path = target
            builder = self.routes[target]
        else:
            path = {v: k for k, v in self.routes.items()}[target]
            builder = target

        async def build() -> None:
            try:
                with self.content:
                    ui.run_javascript(f'''
                        if (window.location.pathname !== "{path}") {{
                            history.pushState({{page: "{path}"}}, "", "{path}");
                        }}
                    ''', timeout=30.0)
                    result = builder()
                    if helpers.is_coroutine_function(builder):
                        await result
            except TimeoutError:
                logger.debug('Router: TimeoutError')
                pass  # the client might have disconnected
        self.content.clear()
        background_tasks.create(build())

    def frame(self) -> ui.element:
        self.content = RouterFrame().on('open', lambda e: self.open(e.args))
        return self.content


router = Router()



import os
import argparse
from .main_ui import run_ui as run_web_ui

parser = argparse.ArgumentParser(
    prog='PySimultanUI',
    description='Run PySimultanUI',
    epilog='')


parser.add_argument("--title", required=False, default='PySimultan Web UI', type=str)
parser.add_argument("--dark", required=False, default=False, type=bool)
parser.add_argument("--reload", required=False, default=False, type=bool)
parser.add_argument("--port", required=False, default=8080, type=int)
parser.add_argument("--uvicorn_logging_level", required=False)
parser.add_argument("--endpoint_documentation", required=False)
parser.add_argument("--storage_secret", required=False)


def run_ui():

    args = parser.parse_args()

    run_web_ui(title=args.title,
               dark=args.dark,
               reload=args.reload,
               port=args.port,
               uvicorn_logging_level=args.uvicorn_logging_level,
               endpoint_documentation=args.endpoint_documentation,
               storage_secret=args.storage_secret if args.storage_secret is not None else
               os.environ.get('STORAGE_SECRET', 'my_secret6849')
               )



from nicegui import ui
from typing import Optional, Union, Any
from ..type_view import TypeView
from nicegui import ui, events, app

from numpy import ndarray
from pandas import DataFrame

from ..component_detail_base_view import ComponentDetailBaseView

from ...core.edit_dialog import DictEditDialog
from ...core.edit_dialog import ContentEditDialog
from ...core.geo_associations import GeoEditDialog

from PySimultan2.simultan_object import SimultanObject
from PySimultan2.default_types import ComponentDictionary
from PySimultan2.files import FileInfo
from PySimultan2.geometry import GeometryModel

from ..parameter_view import ParameterView
from ..mapped_cls.mapped_cls_view import ContentItemView

from SIMULTAN.Data.Components import (SimEnumParameter, SimIntegerParameter, SimStringParameter, SimDoubleParameter)


class DictItemView(object):

    def __init__(self, *args, **kwargs):
        self.component: SimultanObject = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.key = kwargs.get('key')

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):

        from ..detail_views import show_detail

        with ui.item().classes('w-full'):
            with ui.item_section():
                ui.label(f'{self.key}:')

            val = self.component
            if isinstance(val, SimultanObject):
                if val.__ui_element__ is None:
                    self.view_manager.cls_views[val.__class__]['item_view_manager'].add_item_to_view(val)
                with ui.item_section():
                    ui.label(f'{val.name}')
                    ui.button('Details', on_click=lambda e: show_detail(val))
                with ui.item_section():
                    ui.label(f'{val.id}')
            elif isinstance(val, (int, float, str)):
                with ui.item_section():
                    raw_val = self.parent.component.get_raw_attr(self.key)
                    ParameterView(component=val,
                                  raw_val=raw_val,
                                  parent=self).ui_content()
            else:
                with ui.item_section():
                    if hasattr(val, 'name'):
                        ui.label(f'{val.name}:')
                    else:
                        ui.label('No Name')
                with ui.item_section():
                    if hasattr(val, 'id'):
                        ui.label(f'{val.id}:')
                    else:
                        ui.label('No ID')

    def edit(self, event):
        ui.notify('Edit not implemented yet', type='negative')
        raise NotImplementedError

    def remove(self, event):
        ui.notify('Edit not implemented yet', type='negative')
        raise NotImplementedError


class DictView(object):

    def __init__(self, *args, **kwargs):
        self.component: ComponentDictionary = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.card = None

        self.content_item_views: dict[str: ContentItemView] = {}

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def list_content(self):
        with ui.list().classes('w-full'):

            for key, value in self.component.items():
                if self.content_item_views.get(key, None) is None or self.content_item_views[key].component != value:
                    self.content_item_views[key] = DictItemView(component=value,
                                                                parent=self,
                                                                key=key)
                try:
                    self.content_item_views[key].ui_content()
                    ui.separator()
                except Exception as e:
                    print(e)
                    ui.label(f'Could not display {key}: {value}')

        ui.button(f'Add new item to {self.component.name}',
                  on_click=self.parent.add_new_item, icon='add').classes('q-ml-auto')

    @ui.refreshable
    def ui_content(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Content ({len(self.component)})',
                          value=True
                          ).classes('w-full').bind_text_from(self,
                                                                    'data',
                                                                    lambda x: f'Content ({len(self.component)})'
                                                                    ) as self.card:
            self.list_content()


class DictDetailView(ComponentDetailBaseView):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content_view = DictView(component=self.component, parent=self)

    def add_new_item(self, event):
        self.parent.add_new_item(event)

    @property
    def parameters(self) -> tuple[list[tuple[str,
                                             Union[int, float, str],
                                             Union[SimIntegerParameter,
                                                   SimStringParameter,
                                                   SimEnumParameter,
                                                   SimDoubleParameter]
                                             ],
                                  list[tuple[str, SimultanObject]],
                                  list[tuple[str, Union[ndarray, DataFrame]]],
                                  list[tuple[str, FileInfo]],
                                  list[tuple[str, Any]],
                                  list[tuple[str, GeometryModel]],
                                  list[tuple[str, Any]]:
                                       ]
                                  ]:

        parameters = []
        components = []
        arrays = []
        assets = []
        undefined = []
        other = []
        geometry = []

        for key, val in self.component.items():
            if isinstance(val, SimultanObject):
                components.append((key, val))
            elif isinstance(val, (int, float, str)):
                raw_val = self.component.get_raw_attr(key)
                parameters.append((key, val, raw_val))
            elif isinstance(val, (ndarray, DataFrame)):
                arrays.append((key, val))
            elif isinstance(val, FileInfo):
                assets.append((key, val))
            elif val is None:
                undefined.append((key, val))
            else:
                other.append((key, val))

        return parameters, components, arrays, assets, other, geometry, undefined

    @ui.refreshable
    def ui_content(self, *args, **kwargs):
        super().ui_content()
        parameters, components, arrays, assets, other, geometry, undefined = self.parameters
        self.ui_param_table(parameters)
        self.ui_component_table(components)
        self.ui_array_table(arrays)
        self.ui_assets_table(assets)
        self.ui_undefined_table(undefined)

        ui.button(f'Add new item to {self.component.name}',
                  on_click=self.add_new_item, icon='add').classes('q-ml-auto')

    def ui_param_table(self, parameters):
        # create int/float/str table
        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Key', 'field': 'name', 'sortable': True},
                   {'name': 'value', 'label': 'Value', 'field': 'value', 'sortable': True},
                   {'name': 'min', 'label': 'Min', 'field': 'min', 'sortable': True},
                   {'name': 'max', 'label': 'Max', 'field': 'max', 'sortable': True},
                   {'name': 'unit', 'label': 'Unit', 'field': 'unit', 'sortable': True},
                   {'name': 'description', 'label': 'Description', 'field': 'description', 'sortable': False}]

        rows = [{'id': i,
                 'name': param[0],
                 'value': str(param[2].Value),
                 'min': str(param[2].ValueMin) if hasattr(param[2], 'ValueMin') else '',
                 'max': str(param[2].ValueMax) if hasattr(param[2], 'ValueMax') else '',
                 'unit': param[2].Unit if hasattr(param[2], 'Unit') else '',
                 'description': param[2].Description if hasattr(param[2], 'Description') else ''}
                for i, param in enumerate(parameters)]

        param_table = ui.table(columns=columns,
                               rows=rows,
                               title='Parameters',
                               row_key='id').classes('w-full bordered')

        param_table.add_slot('body', r'''
            <q-tr :props="props">
                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                    {{ col.value }}
                </q-td>
                <q-td auto-width>
                    <q-btn size="sm" color="blue" round dense
                           @click="$parent.$emit('edit_val', props)"
                           icon="edit" />
                </q-td>
            </q-tr>
        ''')
        param_table.on('edit_val', self.edit_val)

    def ui_component_table(self, components):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'Key', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'component_id', 'label': 'Component ID', 'field': 'component_id', 'sortable': True},
                   {'name': 'component_name', 'label': 'Component Name', 'field': 'component_name', 'sortable': True},
                   {'name': 'component_type', 'label': 'Component Type', 'field': 'component_type', 'sortable': True},
                   {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True},
                   ]

        rows = [{'id': str(i),
                 'component_id': str(comp[1].id),
                 'name': comp[0],
                 'component_name': comp[1].name,
                 'component_type': comp[1].__class__.__name__,
                 'type': 'Subcomponent' if comp[1]._wrapped_obj in
                                           [y.Component for y in self.component._wrapped_obj.Components]
                 else 'Reference'}
                for i, comp in enumerate(components)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Components',
                              row_key='id').classes('w-full bordered')

        comp_table.add_slot('body', r'''
                    <q-tr :props="props">
                        <q-td v-for="col in props.cols" :key="col.name" :props="props">
                            {{ col.value }}
                        </q-td>
                        <q-td auto-width>
                            <q-btn size="sm" color="blue" round dense
                                   @click="$parent.$emit('detail', props)"
                                   icon="launch" />
                            <q-btn size="sm" color="blue" round dense
                                   @click="$parent.$emit('edit_val', props)"
                                   icon="edit" />
                        </q-td>
                    </q-tr>
                ''')
        comp_table.on('edit_val', self.edit_val)
        comp_table.on('detail', self.show_detail)

    def ui_array_table(self, arrays):
        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'Key', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True}]

        rows = [{'id': str(i),
                 'name': str(comp[0]),
                 'type': 'ND Array' if isinstance(comp[1], ndarray) else 'Table'} for i, comp in enumerate(arrays)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Arrays',
                              row_key='id').classes('w-full bordered')

        comp_table.add_slot('body', r'''
                            <q-tr :props="props">
                                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                    {{ col.value }}
                                </q-td>
                                <q-td auto-width>
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('detail', props)"
                                           icon="launch" />
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('edit_val', props)"
                                           icon="edit" />
                                </q-td>
                            </q-tr>
                        ''')
        comp_table.on('edit_val', self.edit_val)
        comp_table.on('detail', self.show_detail)

    def ui_assets_table(self, assets):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'Key', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'path', 'label': 'Path', 'field': 'path', 'sortable': True}]

        rows = [{'id': i,
                 'name': asset[0],
                 'path': asset[1].path}
                for i, asset in enumerate(assets)]

        asset_table = ui.table(columns=columns,
                               rows=rows,
                               title='Assets',
                               row_key='id').classes('w-full bordered')

        asset_table.add_slot('body', r'''
            <q-tr :props="props">
                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                    {{ col.value }}
                </q-td>
                <q-td auto-width>
                    <q-btn size="sm" color="blue" round dense
                           @click="$parent.$emit('edit_val', props)"
                           icon="edit" />
                </q-td>
            </q-tr>
        ''')
        asset_table.on('edit_val', self.edit_val)

    def ui_undefined_table(self, other):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'Key', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'value', 'label': 'Value', 'field': 'value', 'sortable': True}]

        rows = [{'id': i,
                 'name': param[0],
                 'value': str(param[1])}
                for i, param in enumerate(other)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Undefined Properties',
                              row_key='id').classes('w-full bordered')

        comp_table.add_slot('body', r'''
                            <q-tr :props="props">
                                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                    {{ col.value }}
                                </q-td>
                                <q-td auto-width>
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('edit_val', props)"
                                           icon="edit" />
                                </q-td>
                            </q-tr>
                        ''')
        comp_table.on('edit_val', self.edit_val)

    def edit_val(self, e: events.GenericEventArguments):

        content = next((content for content in self.component._taxonomy_map.content
                        if content.name == e.args['row'].get('name')), None)
        val = getattr(self.component, content.property_name)
        edit_dialog = ContentEditDialog(component=val,
                                        parent=self,
                                        content=content)
        edit_dialog.create_edit_dialog()

    def get_cb_instance(self, e: events.GenericEventArguments):
        key = e.args['row'].get('name')
        instance = getattr(self.component, key)
        if isinstance(instance, (ndarray, )):
            return self.component.get_raw_attr(key).ValueSource.Field
        elif isinstance(instance, DataFrame):
            return self.component.get_raw_attr(key).ValueSource.Table
        else:
            return instance

    def show_detail(self, e: events.GenericEventArguments):
        from ..detail_views import show_detail
        instance = self.get_cb_instance(e)
        show_detail(value=instance)

    def add_new_item(self, event):

        parent = self

        component = self.component
        edit_dialog = DictEditDialog(component=None,
                                     key=None,
                                     parent=self,
                                     options=['Component'])

        def save(self, *args, **kwargs):

            edit_diag = edit_dialog.edit_dialog.edit_dialog

            value = edit_diag.value
            key = edit_dialog.key
            component[key] = value

            if isinstance(value, (int, float)):
                raw_val = component.get_raw_attr(key)
                setattr(raw_val, 'ValueMin', edit_diag.min)
                setattr(raw_val, 'ValueMax', edit_diag.max)
                setattr(raw_val, 'Unit', edit_diag.unit)

            parent.ui_content.refresh()
            edit_dialog.close()

        edit_dialog.save = save
        edit_dialog.create_edit_dialog()


class ComponentDictView(TypeView):

    detail_view = DictDetailView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):
        from ..detail_views import show_detail

        with ui.card().classes(f"{self.colors['item']} w-full") as self.card:
            self.card.on('click', lambda e: show_detail(self.component,
                                                        parent=self.parent)
                         )
            with ui.row().classes(f"{self.colors['item']} w-full") as self.row:
                self.row.on('click', lambda e: show_detail(self.component,
                                                           parent=self.parent)
                            )
                self.checkbox = ui.checkbox(on_change=self.select)
                ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')
                ui.label(f'{str(self.component.id)}')

    def add_new_item(self, event):

        parent = self

        component = self.component
        edit_dialog = DictEditDialog(component=None,
                                     key=None,
                                     parent=self,
                                     options=['Component'])

        def save(self, *args, **kwargs):

            edit_diag = edit_dialog.edit_dialog.edit_dialog

            value = edit_diag.value
            key = edit_dialog.key
            component[key] = value

            if value in [int, float]:
                raw_val = component[key]
                setattr(raw_val, 'ValueMin', edit_diag.min)
                setattr(raw_val, 'ValueMax', edit_diag.max)
                setattr(raw_val, 'Unit', edit_diag.unit)

            parent.content_view.ui_content.refresh()
            edit_dialog.close()

        edit_dialog.save = save
        edit_dialog.create_edit_dialog()



from nicegui import ui, app

from typing import Union, Optional
from .. import user_manager
from ..core.user import User
from PySimultan2.simultan_object import SimultanObject

from PySimultan2.geometry import (GeometryModel, SimultanEdge, SimultanFace,
                                  SimultanVertex, SimultanVolume, SimultanEdgeLoop)


class GeoEditDialog(object):

    def __init__(self, *args, **kwargs) -> None:
        self.parent = kwargs.get('parent')
        self.component: SimultanObject = kwargs.get('component')
        self.dialog = None

        self.geo_model_select = None
        self.geo_type_select = None
        self.geo_select = None

        self.current_options = {}

    @property
    def user(self) -> User:
        return user_manager[app.storage.user['username']]

    @property
    def mapper(self):
        return self.user.mapper

    @property
    def data_model(self):
        return self.user.data_model

    @property
    def geo_models(self):
        return [GeometryModel(wrapped_object=x,
                              object_mapper=self.mapper,
                              data_model=self.data_model) for x in
                self.data_model.models.values()]

    @property
    def selected_geo_model(self):
        return next((geo_model for geo_model in self.geo_models
                     if f'{geo_model.name}; ID: {geo_model.key}' == self.geo_model_select.value),
                    None)

    @property
    def geo_options(self) -> list[any]:

        options = []

        if self.selected_geo_model is None:
            return []

        def generic_options(geo_type: str):
            return [x for x in getattr(self.selected_geo_model, geo_type)]

        def vertex_options():
            return generic_options('vertices')

        def edge_options():
            return generic_options('edges')

        def edge_loop_options():
            return generic_options('edge_loops')

        def face_options():
            return generic_options('faces')

        def volume_options():
            return generic_options('volumes')

        if self.geo_type_select.value == 'All':
            options = [*vertex_options(),
                       *edge_options(),
                       *edge_loop_options(),
                       *face_options(),
                       *volume_options()
                       ]
        elif self.geo_type_select.value == 'Vertex':
            options = vertex_options()
        elif self.geo_type_select.value == 'Edge':
            options = edge_options()
        elif self.geo_type_select.value == 'EdgeLoop':
            options = edge_loop_options()
        elif self.geo_type_select.value == 'Face':
            options = face_options()
        elif self.geo_type_select.value == 'Volume':
            options = volume_options()
        else:
            options = []

        self.current_options = {x.id: x for x in options}

        return options

    @property
    def value(self) -> Union[SimultanVertex, SimultanEdge, SimultanEdgeLoop, SimultanFace, SimultanVolume, None]:
        selected_keys = [int(x['id']) for x in self.geo_select.selected][0]
        return self.current_options.get(selected_keys, None)

    @ui.refreshable
    def ui_content(self):

        self.geo_model_select = ui.select(options=[f'{geo_model.name}; ID: {geo_model.key}'
                                                   for geo_model in self.geo_models],
                                          value=f'{self.geo_models[0].name}; ID: {self.geo_models[0].key}'
                                                if self.geo_models else None,
                                          label='Geometry Model',
                                          on_change=self.on_model_change).classes('w-full')

        self.geo_type_select = ui.select(options=['All', 'Vertex', 'Edge', 'EdgeLoop', 'Face', 'Volume'],
                                         value='All',
                                         label='Geometry Type',
                                         on_change=self.on_type_change).classes('w-full')

        self.geo_ui_content()

    def on_model_change(self, *args, **kwargs):
        self.geo_ui_content.refresh()

    def on_type_change(self, *args, **kwargs):
        self.geo_ui_content.refresh()

    @ui.refreshable
    def geo_ui_content(self):
        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True}]
        rows = [{'id': x.id, 'name': x.name, 'type': type(x).__name__} for x in self.geo_options]
        self.geo_select = ui.table(columns=columns,
                                   rows=rows,
                                   title='Geometry',
                                   selection='single',
                                   pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1}
                                   ).classes('w-full h-full')

    def on_geo_change(self, *args, **kwargs):
        self.geo_ui_content.refresh()

    def create_edit_dialog(self):
        with ui.dialog() as self.dialog, ui.card():
            # ui.label(f'Edit {self.content.name}')
            self.ui_content()
            with ui.row():
                ui.button('Save', on_click=self.save)
                ui.button('Cancel', on_click=self.close)
            self.dialog.open()

    def close(self, *args, **kwargs):
        if self.dialog is not None:
            self.dialog.close()
            self.dialog = None

    def save(self, *args, **kwargs):

        if self.value is None:
            ui.notify('No value selected', type='negative')
            self.close()
            return
        self.value.associate(self.component)
        # self.component.associate = self.value
        ui.notify(f'Associated {self.component.name} with {self.value.name}')
        self.close()
        self.parent.ui_content.refresh()



from nicegui import ui, events, app
# import pdb

from .. import user_manager
from ..core.geo_associations import GeoEditDialog

from PySimultan2.simultan_object import SimultanObject


class ComponentDetailBaseView(object):

    def __init__(self, *args, **kwargs):
        self.component: SimultanObject = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.card = None
        self.row = None

        self.table = None
        self.association_table = None

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @ui.refreshable
    def ui_content(self, *args, **kwargs):
        # with ui.card().classes('w-full h-full'):
        with ui.row().classes('w-full'):
            c_name_input = ui.input(label='Name', value=self.component.name)
            c_name_input.bind_value(self.component, 'name').classes('w-full').style('font-size: 1.25em;')
            with c_name_input.add_slot('append'):
                with ui.button(icon='file_copy', on_click=self.create_copy):
                    ui.tooltip('Create a copy of this component')
        with ui.row():
            ui.label('Global ID: ')
            ui.label(f'{self.component.id.GlobalId.ToString()}')
            ui.label('Local ID: ')
            ui.label(f'{self.component.id.LocalId}')

        self.ui_content_taxonomies()
        self.ui_content_associate_geometry()

    def ui_content_associate_geometry(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Associated Geometry').classes('w-full') as exp:

            columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                       {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                       {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True},
                       {'name': 'actions', 'label': 'Actions', 'field': 'actions', 'sortable': False}]

            rows = [{'id': x.id,
                     'name': x.name,
                     'type': x.__class__.__name__}
                    for x in self.component.associated_geometry]

            self.association_table = ui.table(columns=columns,
                                              rows=rows,
                                              title='Associated Geometry',
                                              pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1},
                                              row_key='id').classes('w-full h-full bordered')

            self.association_table.add_slot('body-cell-actions', r'''
                                                        <q-td key="actions" :props="props">
                                                            <q-btn size="sm" color="blue" round dense
                                                                @click="$parent.$emit('show_detail', props)"
                                                                icon="launch" />
                                                            <q-btn size="sm" color="negative" round dense
                                                                @click="$parent.$emit('delete_association', props)"
                                                                icon="delete" />
                                                        </q-td>
                                                    ''')

            self.association_table.on('show_detail', self.show_geo_detail)
            self.association_table.on('delete_association', self.delete_association)

            with ui.button(icon='add', on_click=self.edit_association):
                ui.tooltip('Add a new association to this component')

    def ui_content_taxonomies(self):

        def add_taxonomy():
            ui.notify('Add taxonomy not implemented yet')

        row = ui.row().classes('w-full')
        with row:
            ui.label('Taxonomies:')
            for taxonomy in self.component.taxonomy_keys:
                ui.chip(taxonomy).classes('m-1')
            with ui.button(icon='add', on_click=add_taxonomy).props('round dense flat justify=end items=center'):
                ui.tooltip('Add a new taxonomy entry to this component')

    def refresh(self):
        self.ui_content.refresh()

    def show_geo_detail(self, e: events.GenericEventArguments):
        from .detail_views import show_detail
        instance = next((x for x in self.component.associated_geometry if x.id == e.args['row'].get('id')), None)._geometry_model
        show_detail(value=instance)

    def edit_association(self):
        edit_dialog = GeoEditDialog(component=self.component,
                                    parent=self)
        edit_dialog.create_edit_dialog()

    def delete_association(self, e: events.GenericEventArguments):
        instance = next((x for x in self.component.associated_geometry if x.id == e.args['row'].get('id')), None)
        instance.disassociate(self.component)
        ui.notify(f'Removed associated {instance.name} from {self.component.name}')
        self.ui_content.refresh()

    def create_copy(self, e: events.GenericEventArguments):
        new_instance = self.component.copy()
        self.user.project_manager.mapped_data.append(new_instance)
        self.user.grid_view.add_item_to_view(new_instance)
        ui.notify(f'Created copy {new_instance.name} {new_instance.id} of {self.component.name}, {self.component.id}')



from PySimultan2 import TaxonomyMap, PythonMapper
from PySimultanUI.src.pysimultanui import MethodMapper, ViewManager

from .classes import Class2, Class3
from .content import contents


mapper = PythonMapper()
method_mapper = MethodMapper()
view_manager = ViewManager()


cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                      taxonomy_key='PySimultan',
                      taxonomy_entry_name='Class2',
                      taxonomy_entry_key='Class2',
                      content=[contents['param_a'],
                               contents['param_b'],
                               contents['param_c'],
                               contents['param_d'],
                               ],
                      )

mapper.register(cls_map.taxonomy_entry_key, Class2, taxonomy_map=cls_map)
mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)

method_mapper.register_method(cls=mapped_cls,
                              name='multiply',
                              method=Class2.multiply)

cls_map2 = TaxonomyMap(taxonomy_name='PySimultan',
                       taxonomy_key='PySimultan',
                       taxonomy_entry_name='Class3',
                       taxonomy_entry_key='Class3',
                       content=[contents['thermal_conductivity'],
                                contents['thermal_resistance'],
                                ],
                       )

mapper.register(cls_map2.taxonomy_entry_key, Class3, taxonomy_map=cls_map2)
mapped_cls2 = mapper.get_mapped_class(cls_map2.taxonomy_entry_key)

method_mapper.register_method(cls=mapped_cls2,
                              name='test_method',
                              method=Class3.test_method)



from nicegui import app
from PySimultan2 import config
from PySimultan2.object_mapper import PythonMapper


def patch(user_manager):

    class GlobalData:
        def __init__(self):
            self._data_model = None
            self._mapper = None
            self._user_manager = user_manager

        @property
        def user(self):
            return self._user_manager.users[app.storage.user['username']]

        @property
        def data_model(self):
            return self._data_model

        @data_model.setter
        def data_model(self, value):
            self._data_model = value

        @property
        def mapper(self):
            if self._mapper is None:
                self._mapper = PythonMapper()
            return self._mapper

        @mapper.setter
        def mapper(self, value):
            self._mapper = value

    global_data = GlobalData()

    def get_user_default_data_model(*args, **kwargs):
        # user_manager.users[app.storage.user['username']]

        try:
            return user_manager.current_user.data_model
        except RuntimeError:
            return None

    def set_user_default_data_model(data_model, *args, **kwargs):
        # user_manager.users[app.storage.user['username']]
        # user_manager.current_user.data_model = data_model
        pass

    def get_user_default_mapper(*args, **kwargs):
        # user_manager.users[app.storage.user['username']]
        return user_manager.current_user.mapper

    def set_user_default_mapper(mapper, *args, **kwargs):
        # user_manager.users[app.storage.user['username']]
        # user_manager.current_user.mapper = mapper
        pass

    config.get_default_data_model = get_user_default_data_model
    config.set_default_data_model = set_user_default_data_model

    config.get_default_mapper = get_user_default_mapper
    config.set_default_mapper = set_user_default_mapper



from PySimultan2 import TaxonomyMap, PythonMapper
from PySimultanUI.src.pysimultanui import MethodMapper, ViewManager

from .classes import Class1
from .content import contents


mapper = PythonMapper()
method_mapper = MethodMapper()
view_manager = ViewManager()


cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                      taxonomy_key='PySimultan',
                      taxonomy_entry_name='Class1',
                      taxonomy_entry_key='Class1',
                      content=[contents['param_a'],
                               contents['param_b'],
                               contents['param_c'],
                               ],
                      )

mapper.register(cls_map.taxonomy_entry_key, Class1, taxonomy_map=cls_map)
mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)

method_mapper.register_method(cls=mapped_cls,
                              name='add',
                              method=Class1.add)



from nicegui.elements.scene import Scene


class ExtendedScene(Scene,
                    component='mesh_scene.js',):

    from .objects3d import Mesh as mesh, Wire as wire, Spline as spline

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from PySimultan2 import Content, TaxonomyMap

from .core.user import UserManager
user_manager = UserManager()

import nicegui.binding
nicegui.binding.MAX_PROPAGATION_TIME = 0.1

from .main_ui import run_ui
from .views.detail_views import DetailView
from .core.mappers import MethodMapper, Mapping, ViewManager


__all__ = ['run_ui', 'user_manager', 'Content', 'TaxonomyMap', 'DetailView', 'MethodMapper', 'Mapping', 'ViewManager']



import os
from nicegui import ui
from ..type_view import TypeView

from PySimultan2.files import FileInfo


extension_lookup_dict = {'.html': 'HTML',
               '.apl': 'APL',
               '.asn': 'ASN.1',
               '.conf': 'Asterisk',
               '.b': 'Brainfuck',
               '.c': 'C',
               '.cs': 'C#',
               '.cpp': 'C++',
               '.hpp': 'C++',
               '.clj': 'Clojure',
               '.cljs': 'ClojureScript',
               '.gss': 'Closure Stylesheets (GSS)',
               'CMakeLists.txt': 'CMake',
               '.cbl': 'Cobol',
               '.coffee': 'CoffeeScript',
               '.lisp': 'Common Lisp',
               '.cql': 'CQL',
               '.cr': 'Crystal',
               '.css': 'CSS',
               '.cypher': 'Cypher',
               '.pyx': 'Cython',
               '.d': 'D',
               '.dart': 'Dart',
               '.diff': 'diff',
               'Dockerfile': 'Dockerfile',
               '.dtd': 'DTD',
               '.dylan': 'Dylan',
               '.ebnf': 'EBNF',
               '.ecl': 'ECL',
               '.edn': 'edn',
               '.e': 'Eiffel',
               '.elm': 'Elm',
               '.erl': 'Erlang',
               '.epl': 'Esper',
               '.fs': 'Forth',
               '.factor': 'Factor',
               '.fcl': 'FCL',
               '.f': 'Fortran',
               '.s': 'Gas',
               '.feature': 'Gherkin',
               '.go': 'Go',
               '.groovy': 'Groovy',
               '.hs': 'Haskell',
               '.hx': 'Haxe',
               '.http': 'HTTP',
               '.hxml': 'HXML',
               '.idl': 'Web IDL',
               '.java': 'Java',
               '.js': 'JavaScript',
               '.jinja2': 'Jinja2',
               '.json': 'JSON',
               '.jsonld': 'JSON-LD',
               '.jsx': 'JSX',
               '.jl': 'Julia',
               '.kt': 'Kotlin',
               '.tex': 'LaTeX',
               '.less': 'LESS',
               '.liquid': 'Liquid',
               '.ls': 'LiveScript',
               '.lua': 'Lua',
               '.sql': 'SQL',
               '.md': 'Markdown',
               '.m': 'Octave',
               '.mbox': 'Mbox',
               '.mrc': 'mIRC',
               '.mo': 'Modelica',
               '.msc': 'MscGen',
               '.msgenny': 'MsGenny',
               '.mps': 'MUMPS',
               '.nginx': 'Nginx',
               '.nsi': 'NSIS',
               '.nt': 'NTriples',
               '.mm': 'Objective-C++',
               '.ml': 'OCaml',
               '.oz': 'Oz',
               '.pas': 'Pascal',
               '.pl': 'Perl',
               '.pgp': 'PGP',
               '.php': 'PHP', '.pig': 'Pig', '.plsql': 'PLSQL', '.ps1': 'PowerShell', '.properties': 'Properties files',
               '.proto': 'ProtoBuf', '.pug': 'Pug', '.pp': 'Puppet', '.py': 'Python', '.q': 'Q', '.r': 'R',
               '.changes': 'RPM Changes', '.spec': 'RPM Spec', '.rb': 'Ruby', '.rs': 'Rust', '.sas': 'SAS', '.sass': 'Sass',
               '.scala': 'Scala', '.scm': 'Scheme', '.scss': 'SCSS', '.sh': 'Shell', '.sieve': 'Sieve',
               '.st': 'Smalltalk', '.sml': 'SML', '.solr': 'Solr', '.sparql': 'SPARQL', '.ods': 'Spreadsheet',
               '.sqlite': 'SQLite', '.nut': 'Squirrel', '.stex': 'sTeX', '.styl': 'Stylus', '.swift': 'Swift',
               '.sv': 'SystemVerilog', '.tcl': 'Tcl', '.textile': 'Textile', '.tiddly': 'TiddlyWiki',
               '.tiki': 'Tiki wiki', '.toml': 'TOML', '.tr': 'Troff', '.tsx': 'TSX', '.ttcn': 'TTCN',
               '.cfg': 'TTCN_CFG', '.ttl': 'Turtle', '.ts': 'TypeScript', '.vb': 'VB.NET',
               '.vbs': 'VBScript', '.vm': 'Velocity', '.v': 'Verilog', '.vhdl': 'VHDL', '.vue': 'Vue',
               '.wasm': 'WebAssembly', '.xml': 'XML', '.xq': 'XQuery', '.xu': 'Xù', '.ys': 'Yacas',
               '.yml': 'YAML',
               '.yaml': 'YAML',
               '.z80': 'Z80'}


class AssetDetailView(object):

    def __init__(self, *args, **kwargs):
        self.component: FileInfo = kwargs.get('component')
        self.parent = kwargs.get('parent')

        self.editor = None

    @ui.refreshable
    def ui_content(self, *args, **kwargs):
        with ui.row():
            ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name').classes('w-full')

        with ui.row():
            ui.label('Key:')
            ui.label(self.component.resource_entry.Key)

        with ui.row():
            ui.label('Size:')
            ui.label(f'{self.component.file_size / 1024:.3f} KB' if self.component.file_size/1024 < 1024
                     else f'{self.component.file_size / 1024 / 1024:.3f} MB')

        with ui.row():
            ui.label('Last Modified:')
            ui.label(str(self.component.last_modified))

        file_extension = os.path.splitext(self.component.name)[1]

        # if file_extension in ['.txt']:
        #     with ui.row():
        #         editor = ui.editor(placeholder='Type something here').bind_value(self.component, 'content',
        #                                                                          backward=lambda x: x)

        try:
            with (ui.card().classes('w-full h-full')):
                content = self.component.content

                file_extension = os.path.splitext(self.component.name)[1]

                self.editor = ui.codemirror(value=content,
                                            language=extension_lookup_dict.get(file_extension, None)
                                            ).classes('w-full h-full')

            with ui.row().classes('w-full'):
                ui.select(self.editor.supported_languages, label='Language', clearable=True) \
                    .classes('w-32').bind_value(self.editor, 'language')
                ui.select(self.editor.supported_themes, label='Theme') \
                    .classes('w-32').bind_value(self.editor, 'theme')

                ui.button('Save', on_click=self.save, icon='save').classes('q-mt-md')

        except Exception as e:
            ui.notify(f'Could not load content: {e}')

    def save(self, event):
        self.component.content = self.editor.value
        ui.notify('Saved!')


class AssetView(TypeView):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def asset(self):
        return self.component

    @ui.refreshable
    def ui_content(self):
        with ui.card().classes('w-full h-full').props('color="blue-800" keep-color') as self.card:
            with ui.item().classes('w-full h-full'):
                with ui.item_section():
                    self.checkbox = ui.checkbox()
                with ui.item_section():
                    ui.label(self.asset.name)
                with ui.item_section():
                    ui.label(str(self.asset.file_size / 1024))
                with ui.item_section():
                    ui.label(str(self.asset.last_modified))
                with ui.item_section():
                    dl_button = ui.button(icon='download', on_click=self.download).classes('q-ml-auto')

    def download(self, event):
        ui.download(f'assets/{self.asset.name}')



import os

from PySimultan2.geometry import GeometryModel
from nicegui import ui, events, app

# from ..detail_views import show_detail
from .tools import create_stl_file
from ..type_view import TypeView
from ...geometry_3d.three_simgeo import display_simgeo
from ...geometry_3d.scene import ExtendedScene
from ... import user_manager

if not os.path.exists('/static/stl'):
    os.makedirs('/static/stl')

app.add_static_files('/stl', '/static/stl')


class GeometryDetailView(object):

    def __init__(self, *args, **kwargs):
        self.component: GeometryModel = kwargs.get('component')
        self.parent: GeometryView = kwargs.get('parent')
        self.file_lookup = None

    @property
    def geometry_model(self) -> GeometryModel:
        return self.component

    def ui_content(self, *args, **kwargs):
        with ui.row():
            ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')

        with ui.row():
            ui.label('Key:')
            ui.label(self.geometry_model.key)

        with ExtendedScene(on_click=self.handle_click).classes('w-full h-full') as self.scene_3d:
            for volume in self.component.volumes:
                display_simgeo(volume, self.scene_3d)
            for face in self.component.faces:
                display_simgeo(face, self.scene_3d)
            for edge_loop in self.component.edge_loops:
                display_simgeo(edge_loop, self.scene_3d)
            for edge in self.component.edges:
                display_simgeo(edge, self.scene_3d)
            for vertex in self.component.vertices:
                display_simgeo(vertex, self.scene_3d)

        ui.button('Show Fullscreen', on_click=self.show_geometry)

        with ui.expansion(icon='format_list_bulleted', text='Geometry').classes('w-full'):

            self.ui_vertices_table()
            self.ui_edges_table()
            self.ui_faces_table()
            self.ui_volumes_table()

        ui.button('Delete', on_click=self.delete_model)

    def ui_vertices_table(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Vertices ({len(self.geometry_model.vertices)})').classes('w-full h-full') as exp:
            columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                       {'name': 'x', 'label': 'X', 'field': 'x', 'sortable': True},
                       {'name': 'y', 'label': 'Y', 'field': 'y', 'sortable': True},
                       {'name': 'z', 'label': 'Z', 'field': 'z', 'sortable': True}]
            rows = [{'id': vertex.id, 'x': vertex.x, 'y': vertex.y, 'z': vertex.z}
                    for vertex in self.geometry_model.vertices]
            ui.table(columns=columns,
                     rows=rows,
                     title='Vertices',
                     pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1}).classes('w-full h-full')

    def ui_edges_table(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Edges ({len(self.geometry_model.edges)})').classes('w-full h-full') as exp:
            columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                       {'name': 'vertex_0', 'label': 'Vertex 0', 'field': 'vertex_0', 'sortable': True},
                       {'name': 'vertex_1', 'label': 'Vertex 1', 'field': 'vertex_1', 'sortable': True},
                       {'name': 'length', 'label': 'Length', 'field': 'length', 'sortable': True}]
            rows = [{'id': edge.id, 'vertex_0': edge.vertex_0.id, 'vertex_1': edge.vertex_1.id, 'length': edge.length}
                    for edge in self.geometry_model.edges]
            ui.table(columns=columns,
                     rows=rows,
                     title='Edges',
                     pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1}).classes('w-full h-full')

    def ui_faces_table(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Faces ({len(self.geometry_model.faces)})').classes('w-full h-full') as exp:
            columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                       {'name': 'area', 'label': 'Area', 'field': 'area', 'sortable': True}]
            rows = [{'id': face.id, 'area': face.area}
                    for face in self.geometry_model.faces]
            ui.table(columns=columns,
                     rows=rows,
                     title='Faces',
                     pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1}).classes('w-full h-full')

    def ui_volumes_table(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Volumes ({len(self.geometry_model.volumes)})').classes('w-full h-full') as exp:
            columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                       {'name': 'volume', 'label': 'Volume', 'field': 'volume', 'sortable': True}]
            rows = [{'id': volume.id, 'volume': volume.volume}
                    for volume in self.geometry_model.volumes]
            ui.table(columns=columns,
                     rows=rows,
                     title='Volumes',
                     pagination={'rowsPerPage': 5, 'sortBy': 'id', 'page': 1}).classes('w-full h-full')

    def delete_model(self, *args, **kwargs):
        ui.notify('Delete not implemented yet', type='negative')

    def show_geometry(self):
        # create dialog with fc_geometry
        with ui.dialog() as dialog, ui.card().classes('w-full h-full'):

            if self.file_lookup is None:
                self.file_lookup = create_stl_file(self.geometry_model)
            with ui.scene(on_click=self.handle_click).classes('w-full h-full') as self.scene_3d:
                self.scene_3d.spot_light(distance=100, intensity=0.5).move(-10, 0, 10)
                for f_id, f in self.file_lookup.items():
                    self.scene_3d.stl(f[0]).with_name(str(f_id)).material(f[1])

            ui.button('Cancel', on_click=dialog.close)

        dialog.open()

    def handle_click(self, e: events.SceneClickEventArguments, *args, **kwargs):
        if e.hits:
            hit = e.hits[0]
            name = hit.object_name or hit.object_id
            ui.notify(f'You clicked on the {name} at ({hit.x:.2f}, {hit.y:.2f}, {hit.z:.2f})')

    def refresh(self):
        self.ui_content()


class GeometryView(TypeView):

    detail_view = GeometryDetailView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def geometry_model(self) -> GeometryModel:
        return self.component

    @ui.refreshable
    def ui_content(self):
        from ..detail_views import show_detail
        with ui.card().classes(f"{self.colors['item']} w-full h-full") as self.card:
            self.card.on('click', lambda e: show_detail(value=self.component,
                                                        parent=self)
                         )
            with ui.row().classes('h-full w-full') as self.row:
                self.row.on('click', lambda e: show_detail(value=self.component,
                                                           parent=self)
                            )
                self.checkbox = ui.checkbox(on_change=self.select)
                ui.label('Name:')
                ui.label(self.geometry_model.name)
                ui.label('Key:')
                ui.label(self.geometry_model.key)

    def delete_model(self, *args, **kwargs):
        ui.notify('Delete not implemented yet', type='negative')
